from typing import Dict
import numpy as np  # type: ignore


class RegistroEstado:
    """
    Armazena as informações de um registro visitado pelo NEWAVE
    em algum período de alguma simulação.

    ** Parâmetros **

    - ireg: `int`
    - itec: `int`
    - simc: `int`
    - itef: `int`
    - fobj: `int`
    - tabela: `np.ndarray`

    """

    __slots__ = ["ireg",
                 "itec",
                 "simc",
                 "itef",
                 "fobj",
                 "tabela"]

    def __init__(self,
                 ireg: int,
                 itec: int,
                 simc: int,
                 itef: int,
                 fobj: float,
                 tabela: np.ndarray):
        self.ireg = ireg
        self.itec = itec
        self.simc = simc
        self.itef = itef
        self.fobj = fobj
        self.tabela = tabela

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre RegistroEstado avalia todos os
        valores.
        """
        if not isinstance(o, RegistroEstado):
            return False
        reg: RegistroEstado = o
        eq_ireg = self.ireg == reg.ireg
        eq_itec = self.itec == reg.itec
        eq_simc = self.simc == reg.simc
        eq_itef = self.itef == reg.itef
        eq_fobj = self.fobj == reg.fobj
        eq_tab = np.array_equal(self.tabela, reg.tabela)

        return all([eq_ireg,
                    eq_itec,
                    eq_simc,
                    eq_itef,
                    eq_fobj,
                    eq_tab])


class Estados:
    """
    Armazena os dados dos estados visitados pelo NEWAVE existentes
    no arquivo `estados.rel` do NWLISTCF.

    Esta classe armazena os estados de cada uma das variáveis envolvidas
    no problema e da função objetivo, para cada registro e REE dentro
    do registro.

    Cada registro possui um modelo próprio, armazenando os estados das
    variáveis em uma array específica.

    **Parâmetros**

    - registros: `Dict[int, Dict[int, RegistroEstado]]`

    """
    def __init__(self,
                 registros: Dict[int, Dict[int, RegistroEstado]]):
        self.registros = registros

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre Estados avalia todos os registros.
        """
        if not isinstance(o, Estados):
            return False
        est: Estados = o
        if self.registros.keys() != est.registros.keys():
            return False
        for (p1, p2) in zip(self.registros.values(),
                            est.registros.values()):
            if p1.keys() != p2.keys():
                return False
            for (r1, r2) in zip(p1.values(),
                                p2.values()):
                if r1 != r2:
                    return False
        return True



# Imports do próprio módulo
from inewave._utils.escrita import Escrita
from inewave._utils.leitura import Leitura
from .modelos.dger import DGer
from .modelos.dger import EnumImpressaoConvergencia
from .modelos.dger import EnumImpressaoOperacao
from .modelos.dger import EnumTamanhoArquivoVazoes
from .modelos.dger import EnumDuracaoPatamar
from .modelos.dger import EnumCorrecaoEnergiaDesvio
from .modelos.dger import EnumInicioTesteConvergencia
from .modelos.dger import EnumMatrizCorrelacaoEspacial
from .modelos.dger import EnumMomentoReamostragem
from .modelos.dger import EnumRepresentacaoSubmotorizacao
from .modelos.dger import EnumRepresentanteAgregacao
from .modelos.dger import EnumTendenciaHidrologica
from .modelos.dger import EnumTipoExecucao
from .modelos.dger import EnumTipoGeracaoENAs
from .modelos.dger import EnumTipoReamostragem
from .modelos.dger import EnumTipoSimulacaoFinal
from .modelos.dger import EnumSazonaliza
from .modelos.dger import EnumRepresentacaoIncerteza
# Imports de módulos externos
import os
from traceback import print_exc


class LeituraDGer(Leitura):
    """
    Realiza a leitura do arquivo dger.dat
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo dger.dat, construindo
    um objeto `DGer` cujas informações são as mesmas do dger.dat.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.

    Uma vez realizada a leitura do arquivo, as informações são guardadas
    internamente no atributo `dger`.

    **Exemplos**

    >>> diretorio = "~/documentos/.../deck"
    >>> leitor = LeituraDGer(diretorio)
    >>> leitor.le_arquivo()
    # Ops, esqueci de pegar o objeto
    >>> dger = leitor.dger

    """
    def __init__(self,
                 diretorio: str) -> None:
        super().__init__()
        self.diretorio = diretorio
        # DGer default, depois é substituído
        self.dger = DGer.dger_padrao()

    def le_arquivo(self) -> DGer:
        """
        Realiza a leitura do arquivo `dger.dat`.
        """
        try:
            caminho = os.path.join(self.diretorio, "dger.dat")
            with open(caminho, "r") as arq:
                # Lê o nome do estudo e restringe até a coluna 80
                nome = self._le_linha_com_backup(arq)
                self.dger.nome_estudo = nome[:min([79, len(nome)])]
                # Lê os demais parâmetros
                ci = 21
                cf = 25

                def le_parametro(linha: str):
                    aux = linha[:21]
                    param = linha[ci:cf].strip()
                    if "TIPO DE EXECUCAO" in aux:
                        t = EnumTipoExecucao.infere_valor(int(param))
                        self.dger.tipo_execucao = t
                    elif "DURACAO DO PERIODO" in aux:
                        self.dger.duracao_estagio_op = int(param)
                    elif "No. DE ANOS DO EST" in aux:
                        self.dger.num_anos_estudo = int(param)
                    elif "MES INICIO PRE-EST" in aux:
                        self.dger.mes_inicio_pre_estudo = int(param)
                    elif "MES INICIO DO ESTUDO" in aux:
                        self.dger.mes_inicio_estudo = int(param)
                    elif "ANO INICIO DO ESTUDO" in aux:
                        self.dger.ano_inicio_estudo = int(param)
                    elif "No. DE ANOS PRE" in aux:
                        self.dger.num_anos_pre_estudo = int(param)
                    elif "No. DE ANOS POS FINAL" in aux:
                        self.dger.num_anos_pos_sim_final = int(param)
                    elif ("No. DE ANOS POS" in aux and "FINAL" not in aux):
                        self.dger.num_anos_pos_estudo = int(param)
                    elif "IMPRIME DADOS" in aux:
                        p = param
                        self.dger.imprime_dados_usinas = (True if p == "1"
                                                          else False)
                    elif "IMPRIME MERCADOS" in aux:
                        p = param
                        self.dger.imprime_dados_mercados = (True if p == "1"
                                                            else False)
                    elif "IMPRIME ENERGIAS" in aux:
                        p = param
                        self.dger.imprime_dados_energias = (True if p == "1"
                                                            else False)
                    elif "IMPRIME M. ESTOCAS" in aux:
                        p = param
                        self.dger.imprime_dados_modelo = (True if p == "1"
                                                          else False)
                    elif "IMPRIME SUBSISTEMA" in aux:
                        p = param
                        self.dger.imprime_dados_rees = (True if p == "1"
                                                        else False)
                    elif "No MAX. DE ITER." in aux:
                        self.dger.max_iteracoes = int(param)
                    elif "No DE SIM. FORWARD" in aux:
                        self.dger.num_sim_forward = int(param)
                    elif "No DE ABERTURAS" in aux:
                        self.dger.num_aberturas = int(param)
                    elif "No DE SERIES SINT." in aux:
                        self.dger.num_series_sinteticas = int(param)
                    elif "ORDEM MAX. PAR(P)" in aux:
                        self.dger.ordem_maxima_parp = int(param)
                    elif "ANO INICIAL HIST." in aux:
                        self.dger.ano_inicial_vaz_historicas = int(param)
                        t = int(linha[28])
                        tam = EnumTamanhoArquivoVazoes.infere_valor(t)
                        self.dger.tamanho_arq_vaz_historicas = tam
                    elif "CALCULA VOL.INICIAL" in aux:
                        p = param
                        self.dger.calcula_vol_inicial = (True if p == "1"
                                                         else False)
                    elif "VOLUME INICIAL  -%" in aux:
                        pass
                    elif "POR SUBSISTEMA" in aux:
                        cisub = 22
                        n_col_sub = 5
                        for i in range(4):
                            cfsub = cisub + n_col_sub
                            v = float(linha[cisub:cfsub])
                            self.dger.vol_inicial_subsistema[i] = v
                            cisub = cfsub + 2
                    elif "TOLERANCIA      -%" in aux:
                        self.dger.tolerancia = float(param)
                    elif "TAXA DE DESCONTO-%" in aux:
                        self.dger.taxa_de_desconto = float(param)
                    elif "TIPO SIMUL. FINAL" in aux:
                        val = int(param)
                        tipo = EnumTipoSimulacaoFinal.infere_valor(val)
                        self.dger.tipo_simulacao_final = tipo
                    elif "IMPRESSAO DA OPER" in aux:
                        val = int(param)
                        impr_op = EnumImpressaoOperacao.infere_valor(val)
                        self.dger.impressao_operacao = impr_op
                    elif "IMPRESSAO DA CONVERG." in aux:
                        val = int(param)
                        impr_c = EnumImpressaoConvergencia.infere_valor(val)
                        self.dger.impressao_convergencia = impr_c
                    elif "INTERVALO P/ GRAVAR" in aux:
                        self.dger.intervalo_gravacao_relatorio = int(param)
                    elif "No. MIN. ITER." in aux:
                        self.dger.min_interacoes = int(param)
                    elif "RACIONAMENTO PREVENT." in aux:
                        p = param
                        self.dger.racionamento_preventivo = (True if p == "1"
                                                             else False)
                    elif "No. ANOS MANUT.UTE'S" in aux:
                        self.dger.numero_anos_manutencao_UTEs = int(param)
                    elif "TENDENCIA HIDROLOGICA" in aux:
                        val = int(param)
                        tend = EnumTendenciaHidrologica.infere_valor(val)
                        self.dger.tendencia_hidrologica = tend
                    elif "RESTRICA0 DE ITAIPU" in aux:
                        p = param
                        self.dger.restricoes_itaipu = (True if p == "1"
                                                       else False)
                    elif "BID " in aux:
                        p = param
                        self.dger.bidding_demanda = (True if p == "1"
                                                     else False)
                    elif "PERDAS P/ TRANSMISSAO" in aux:
                        p = param
                        self.dger.perdas_transmissao = (True if p == "1"
                                                        else False)
                    elif "EL NINO" in aux:
                        p = param
                        self.dger.el_nino = (True if p == "1"
                                             else False)
                    elif "ENSO INDEX " in aux:
                        p = param
                        self.dger.enso = (True if p == "1"
                                          else False)
                    elif "DURACAO POR PATAMAR" in aux:
                        val = int(param)
                        dp = EnumDuracaoPatamar.infere_valor(val)
                        self.dger.duracao_patamar = dp
                    elif "OUTROS USOS DA AGUA" in aux:
                        p = param
                        self.dger.considera_desvio_dagua = (True if p == "1"
                                                            else False)
                    elif "CORRECAO DESVIO" in aux:
                        val = int(param)
                        des = EnumCorrecaoEnergiaDesvio.infere_valor(val)
                        self.dger.correcao_energia_desvio = des
                    elif "C.AVERSAO/PENAL.VMINP" in aux:
                        p = param
                        self.dger.considera_curva_aversao = (True if p == "1"
                                                             else False)
                    elif "TIPO DE GERACAO ENAS" in aux:
                        val = int(param)
                        tg = EnumTipoGeracaoENAs.infere_valor(val)
                        self.dger.tipo_geracao_afluencias = tg
                    elif "RISCO DE DEFICIT" in aux:
                        r1 = float(linha[21:25].strip())
                        r2 = float(linha[27:31].strip())
                        self.dger.profundidade_risco_deficit = (r1, r2)
                    elif "ITERACAO P/SIM.FINAL" in aux:
                        self.dger.iteracao_sim_final = int(param)
                    elif "AGRUPAMENTO LIVRE" in aux:
                        p = param
                        self.dger.agrupamento_livre_interc = (True
                                                              if p == "1"
                                                              else False)
                    elif "EQUALIZACAO PEN.INT." in aux:
                        p = param
                        self.dger.equaliza_penalidades_interc = (True
                                                                 if p == "1"
                                                                 else False)
                    elif "REPRESENT.SUBMOT." in aux:
                        t = int(param)
                        sm = EnumRepresentacaoSubmotorizacao.infere_valor(t)
                        self.dger.representa_submotor = sm
                    elif "ORDENACAO AUTOMATICA" in aux:
                        p = param
                        self.dger.ordenacao_automatica_subsist = (True
                                                                  if p == "1"
                                                                  else False)
                    elif "CONS. CARGA ADICIONAL" in aux:
                        p = param
                        self.dger.considera_cargas_adicionais = (True
                                                                 if p == "1"
                                                                 else False)
                    elif "DELTA ZSUP" in aux:
                        self.dger.delta_zsup = float(param)
                    elif "DELTA ZINF" in aux:
                        self.dger.delta_zinf = float(param)
                    elif "DELTAS CONSECUT." in aux:
                        self.dger.deltas_consecutivos = int(param)
                    elif "DESP. ANTEC.  GNL" in aux:
                        p = param
                        self.dger.considera_despacho_gnl = (True
                                                            if p == "1"
                                                            else False)
                    elif "MODIF.AUTOM.ADTERM" in aux:
                        p = param
                        self.dger.modifica_auto_despacho_gnl = (True
                                                                if p == "1"
                                                                else False)
                    elif "CONSIDERA GHMIN" in aux:
                        p = param
                        self.dger.considera_ghmin = (True
                                                     if p == "1"
                                                     else False)
                    elif "S.F. COM DATA" in aux:
                        p = param
                        self.dger.sim_final_com_data = (True
                                                        if p == "1"
                                                        else False)
                    elif "GER.PLs E NV1 E NV2" in aux:
                        p1 = linha[21:25].strip()
                        p2 = linha[26:30].strip()
                        p3 = linha[31:35].strip()
                        p4 = linha[36:40].strip()
                        p5 = linha[41:45].strip()
                        self.dger.gerenciador_externo = (True if p1 == "1"
                                                         else False)
                        self.dger.comunicacao_dois_niveis = (True if p2 == "1"
                                                             else False)
                        self.dger.armazenamento_local_temp = (True if p3 == "1"
                                                              else False)
                        self.dger.aloca_memoria_enas = (True if p4 == "1"
                                                        else False)
                        self.dger.aloca_memoria_cortes = (True if p5 == "1"
                                                          else False)
                    elif "SAR" in aux:
                        p = param
                        self.dger.sar = (True if p == "1"
                                         else False)
                    elif "CVAR" in aux:
                        p = param
                        self.dger.cvar = (True if p == "1"
                                          else False)
                    elif "CONS. ZSUP MIN. CONV." in aux:
                        p = param
                        self.dger.convergencia_minimo_zsup = (True if p == "1"
                                                              else False)
                    elif "DESCONSIDERA VAZMIN" in aux:
                        p = param
                        self.dger.desconsidera_vazao_minima = (True if p == "1"
                                                               else False)
                    elif "RESTRICOES ELETRICAS" in aux:
                        p = param
                        self.dger.considera_restricoes_elet = (True if p == "1"
                                                               else False)
                    elif "SELECAO DE CORTES" in aux:
                        p = param
                        self.dger.selecao_cortes_benders = (True if p == "1"
                                                            else False)
                    elif "JANELA DE CORTES" in aux:
                        p = param
                        self.dger.janela_selecao_cortes = (True if p == "1"
                                                           else False)
                    elif "REAMOST. CENARIOS" in aux:
                        reamos = linha[21:25].strip()
                        self.dger.reamostragem = bool(int(reamos))
                        reamos = linha[26:30].strip()
                        t = int(reamos)
                        tipo_reamos = EnumTipoReamostragem.infere_valor(t)
                        self.dger.tipo_reamostragem = tipo_reamos
                        self.dger.passo_reamostragem = int(linha[31:35])
                    elif "CONVERGE NO ZERO" in aux:
                        p = param
                        self.dger.considera_convergencia_no0 = (True
                                                                if p == "1"
                                                                else False)
                    elif "CONSULTA FCF" in aux:
                        p = param
                        self.dger.consulta_fcf = (True
                                                  if p == "1"
                                                  else False)
                    elif "IMPRESSAO ENA" in aux:
                        self.dger.impressao_ena = bool(int(param))
                    elif "IMP. CATIVO S.FINAL" in aux:
                        self.dger.impressao_cortes_ativos = bool(int(param))
                    elif "REP. AGREGACAO" in aux:
                        t = int(param)
                        ra = EnumRepresentanteAgregacao.infere_valor(t)
                        self.dger.representante_agregacao = ra
                    elif "MATRIZ CORR.ESPACIAL" in aux:
                        t = int(param)
                        mc = EnumMatrizCorrelacaoEspacial.infere_valor(t)
                        self.dger.matriz_corr_espacial = mc
                    elif "DESCONS. CONV. ESTAT" in aux:
                        p = param
                        self.dger.desconsidera_converg_estatist = bool(int(p))
                    elif "MOMENTO REAMOSTRAGEM" in aux:
                        t = int(param)
                        mr = EnumMomentoReamostragem.infere_valor(t)
                        self.dger.momento_reamostragem = mr
                    elif "ARQUIVOS ENA" in aux:
                        p = param
                        self.dger.mantem_arquivos_ena = bool(int(p))
                    elif "INICIO TESTE CONVERG." in aux:
                        t = int(param)
                        tc = EnumInicioTesteConvergencia.infere_valor(t)
                        self.dger.inicio_teste_convergencia = tc
                    elif "SAZ. VMINT PER. EST." in aux:
                        t = int(param)
                        sv = EnumSazonaliza.infere_valor(t)
                        self.dger.sazonaliza_vmint = sv
                    elif "SAZ. VMAXT PER. EST." in aux:
                        t = int(param)
                        sv = EnumSazonaliza.infere_valor(t)
                        self.dger.sazonaliza_vmaxt = sv
                    elif "SAZ. VMINP PER. EST." in aux:
                        t = int(param)
                        sv = EnumSazonaliza.infere_valor(t)
                        self.dger.sazonaliza_vminp = sv
                    elif "SAZ. CFUGA e CMONT" in aux:
                        t = int(param)
                        sv = EnumSazonaliza.infere_valor(t)
                        self.dger.sazonaliza_cfuga_cmont = sv
                    elif "REST. EMISSAO GEE" in aux:
                        p = param
                        self.dger.restricoes_gee = bool(int(p))
                    elif "AFLUENCIA ANUAL PARP" in aux:
                        af1 = bool(int(linha[21:25].strip()))
                        af2 = bool(int(linha[26:30].strip()))
                        self.dger.afluencia_anual_parp = (af1, not af2)
                    elif "INCERTEZA GER.EOLICA" in aux:
                        p = param
                        self.dger.incerteza_ger_eolica = bool(int(p))
                    elif "INCERTEZA GER.SOLAR" in aux:
                        p = param
                        self.dger.incerteza_ger_solar = bool(int(p))
                    elif "REPRESENTACAO INCERT" in aux:
                        t = int(param)
                        ri = EnumRepresentacaoIncerteza.infere_valor(t)
                        self.dger.representacao_incerteza = ri
                    elif "REST. FORNEC. GAS" in aux:
                        p = param
                        self.dger.restricoes_fornecimento_gas = bool(int(p))

                while True:
                    # Confere se o arquivo já acabou
                    linha = self._le_linha_com_backup(arq)
                    if len(linha) < 2:
                        break
                    # Senão, lê mais um parâmetro
                    le_parametro(linha)
                return self.dger
        except Exception:
            print_exc()
            return self.dger

    def _fim_arquivo(self, linha: str) -> bool:
        return False


class EscritaDGer(Escrita):
    """
    Realiza a escrita do arquivo dger.dat
    em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para escrever os campos
    de um arquivo dger.dat, a partir de um objeto `DGer`.

    Este objeto existe para retirar do modelo de dados a complexidade
    de armazenar as strings auxiliares do arquivo, desenhar tabelas, dentre
    outras tarefas associadas à escrita.

    Se o diretório de escrita não existir, ele será criado.

    **Exemplos**

    >>> diretorio = "~/documentos/.../deck"
    >>> dger = DGer.dger_padrao()
    >>> escritor = EscritaDGer(diretorio)
    >>> escritor.escreve_arquivo(dger)
    """
    def __init__(self, diretorio: str):
        super().__init__()
        self.diretorio = diretorio

    def escreve_arquivo(self, dger: DGer):
        """
        Realiza a escrita de um arquivo `dger.dat`.
        """
        # Confere se o diretório existe. Senão, cria.
        if not os.path.exists(self.diretorio):
            os.makedirs(self.diretorio)
        # Inicia a escrita
        caminho = os.path.join(self.diretorio, "dger.dat")
        len_aux = 21
        len_dado = 4
        with open(caminho, "w") as arq:

            def escreve_alinhado(aux: str,
                                 dado):
                str_aux = aux.ljust(len_aux)
                str_dado = str(int(dado)).rjust(len_dado)
                arq.write(str_aux + str_dado + "\n")

            def escreve_alinhado_f(aux: str,
                                   dado):
                str_aux = aux.ljust(len_aux)
                alinhado = "{:.1f}".format(float(dado)).rjust(4)
                str_dado = str(alinhado).rjust(len_dado)
                arq.write(str_aux + str_dado + "\n")

            # Nome do estudo
            if "\n" in dger.nome_estudo:
                arq.write(dger.nome_estudo)
            else:
                arq.write(dger.nome_estudo + "\n")
            # Tipo de execução
            escreve_alinhado("TIPO DE EXECUCAO",
                             dger.tipo_execucao.value)
            # Duração do período
            escreve_alinhado("DURACAO DO PERIODO",
                             dger.duracao_estagio_op)
            # Número de anos do estudo
            escreve_alinhado("No. DE ANOS DO EST",
                             dger.num_anos_estudo)
            # Mês de início do pré-estudo
            escreve_alinhado("MES INICIO PRE-EST",
                             dger.mes_inicio_pre_estudo)
            # Mês de início do estudo
            escreve_alinhado("MES INICIO DO ESTUDO",
                             dger.mes_inicio_estudo)
            # Ano de início do estudo
            escreve_alinhado("ANO INICIO DO ESTUDO",
                             dger.ano_inicio_estudo)
            # Número de anos pré estudo
            escreve_alinhado("No. DE ANOS PRE",
                             dger.num_anos_pre_estudo)
            # Número de anos pós-estudo
            escreve_alinhado("No. DE ANOS POS",
                             dger.num_anos_pos_estudo)
            # Número de anos pós na sim. final
            escreve_alinhado("No. DE ANOS POS FINAL",
                             dger.num_anos_pos_sim_final)
            # Imprime dados de usinas
            escreve_alinhado("IMPRIME DADOS",
                             dger.imprime_dados_usinas)
            # Imprime dados de mercados
            escreve_alinhado("IMPRIME MERCADOS",
                             dger.imprime_dados_mercados)
            # Imprime energias
            escreve_alinhado("IMPRIME ENERGIAS",
                             dger.imprime_dados_energias)
            # Imprime dados do modelo
            escreve_alinhado("IMPRIME M. ESTOCAS",
                             dger.imprime_dados_modelo)
            # Imprime dados REEs
            escreve_alinhado("IMPRIME SUBSISTEMA",
                             dger.imprime_dados_rees)
            # Máximo de iterações
            escreve_alinhado("No MAX. DE ITER.",
                             dger.max_iteracoes)
            # Número de forwards
            escreve_alinhado("No DE SIM. FORWARD",
                             dger.num_sim_forward)
            # Número de aberturas
            escreve_alinhado("No DE ABERTURAS",
                             dger.num_aberturas)
            # Número de séries sintéticas
            escreve_alinhado("No DE SERIES SINT.",
                             dger.num_series_sinteticas)
            # Ordem máxima PAR(p)
            escreve_alinhado("ORDEM MAX. PAR(P)",
                             dger.ordem_maxima_parp)
            # Ano inicial do histórico
            str_aux = "ANO INICIAL HIST.".ljust(len_aux)
            ano = dger.ano_inicial_vaz_historicas
            str_dado1 = str(int(ano)).rjust(len_dado)
            tam = dger.tamanho_arq_vaz_historicas.value
            str_dado2 = str(int(tam)).rjust(len_dado)
            arq.write(str_aux + str_dado1 + str_dado2 + "\n")
            # Calcula volume inicial
            escreve_alinhado("CALCULA VOL.INICIAL",
                             dger.calcula_vol_inicial)
            # Linha de cabeçalho
            lin = "VOLUME INICIAL  -%   XXX.X  XXX.X  XXX.X  XXX.X  XXX.X"
            arq.write(lin + "\n")
            # Volumes por subsistema
            str_aux = " POR SUBSISTEMA".ljust(len_aux)
            str_dados = ""
            for v in dger.vol_inicial_subsistema:
                str_dados += "{:.1f}".format(v).rjust(5) + "  "
            arq.write(str_aux + str_dados + "\n")
            # Tolerância
            escreve_alinhado_f("TOLERANCIA      -%",
                               dger.tolerancia)
            # Taxa de desconto
            escreve_alinhado_f("TAXA DE DESCONTO-%",
                               dger.taxa_de_desconto)
            # Tipo de sim final
            escreve_alinhado("TIPO SIMUL. FINAL",
                             dger.tipo_simulacao_final.value)
            # Imprime operação
            escreve_alinhado("IMPRESSAO DA OPER",
                             dger.impressao_operacao.value)
            # Imprime convergência
            escreve_alinhado("IMPRESSAO DA CONVERG.",
                             dger.impressao_convergencia.value)
            # Intervalo gravação
            escreve_alinhado("INTERVALO P/ GRAVAR",
                             dger.intervalo_gravacao_relatorio)
            # Min. iterações
            escreve_alinhado("No. MIN. ITER.",
                             dger.min_interacoes)
            # Racionamento preventivo
            escreve_alinhado("RACIONAMENTO PREVENT.",
                             dger.racionamento_preventivo)
            # Num. anos manutenção UTEs
            escreve_alinhado("No. ANOS MANUT.UTE'S",
                             dger.numero_anos_manutencao_UTEs)
            # Tendência hidrológica
            escreve_alinhado("TENDENCIA HIDROLOGICA",
                             dger.tendencia_hidrologica.value)
            # Restrição de Itaipu
            escreve_alinhado("RESTRICA0 DE ITAIPU",
                             dger.restricoes_itaipu)
            # Bidding
            escreve_alinhado("BID",
                             dger.bidding_demanda)
            # Perdas na transmissão
            escreve_alinhado("PERDAS P/ TRANSMISSAO",
                             dger.perdas_transmissao)
            # El Niño
            escreve_alinhado("EL NINO",
                             dger.el_nino)
            # ENSO
            escreve_alinhado("ENSO INDEX",
                             dger.enso)
            # Duração por patamar
            escreve_alinhado("DURACAO POR PATAMAR",
                             dger.duracao_patamar.value)
            # Considera desvio d'água
            escreve_alinhado("OUTROS USOS DA AGUA",
                             dger.considera_desvio_dagua)
            # Correção do desvio
            escreve_alinhado("CORRECAO DESVIO",
                             dger.correcao_energia_desvio.value)
            # Considera curva VminP
            escreve_alinhado("C.AVERSAO/PENAL.VMINP",
                             dger.considera_curva_aversao)
            # Tipo de geração de ENAs
            escreve_alinhado("TIPO DE GERACAO ENAS",
                             dger.tipo_geracao_afluencias.value)
            # Profundidade do risco de déficit
            str_aux = "RISCO DE DEFICIT".ljust(len_aux)
            str_dados = ""
            for v in dger.profundidade_risco_deficit:
                str_dados += "{:.1f}".format(v).rjust(4) + "  "
            arq.write(str_aux + str_dados + "\n")
            # Iteração para simulação final
            escreve_alinhado("ITERACAO P/SIM.FINAL",
                             dger.iteracao_sim_final)
            # Agrupamento livre de intercâmbios
            escreve_alinhado("AGRUPAMENTO LIVRE",
                             dger.agrupamento_livre_interc)
            # Equalização da penalização de intercâmbios
            escreve_alinhado("EQUALIZACAO PEN.INT.",
                             dger.equaliza_penalidades_interc)
            # Representação da submotorização
            escreve_alinhado("REPRESENT.SUBMOT.",
                             dger.representa_submotor.value)
            # Ordenação automática de subsistemas
            escreve_alinhado("ORDENACAO AUTOMATICA",
                             dger.ordenacao_automatica_subsist)
            # Considera cargas adicionais
            escreve_alinhado("CONS. CARGA ADICIONAL",
                             dger.considera_cargas_adicionais)
            # Delta Zsup
            escreve_alinhado_f("DELTA ZSUP",
                               dger.delta_zsup)
            # Delta Zinf
            escreve_alinhado_f("DELTA ZINF",
                               dger.delta_zinf)
            # Deltas consecutivos
            escreve_alinhado("DELTAS CONSECUT.",
                             dger.deltas_consecutivos)
            # Despacho antecipado GNL
            escreve_alinhado("DESP. ANTEC.  GNL",
                             dger.considera_despacho_gnl)
            # Modifica automaticamente AdTerm
            escreve_alinhado("MODIF.AUTOM.ADTERM",
                             dger.modifica_auto_despacho_gnl)
            # Considera GHmin
            escreve_alinhado("CONSIDERA GHMIN",
                             dger.considera_ghmin)
            # Simulação final com data
            escreve_alinhado("S.F. COM DATA",
                             dger.sim_final_com_data)
            # Gerenciados externo, etc...
            str_aux = "GER.PLs E NV1 E NV2".ljust(len_aux)
            str_dados = ""
            d = int(dger.gerenciador_externo)
            str_dados += str(d).rjust(4) + " "
            d = int(dger.comunicacao_dois_niveis)
            str_dados += str(d).rjust(4) + " "
            d = int(dger.armazenamento_local_temp)
            str_dados += str(d).rjust(4) + " "
            d = int(dger.aloca_memoria_enas)
            str_dados += str(d).rjust(4) + " "
            d = int(dger.aloca_memoria_cortes)
            str_dados += str(d).rjust(4) + " "
            arq.write(str_aux + str_dados + "\n")
            # SAR
            escreve_alinhado("SAR",
                             dger.sar)
            # CVaR
            escreve_alinhado("CVAR",
                             dger.cvar)
            # Critério de min Zsup p/ convergência
            escreve_alinhado("CONS. ZSUP MIN. CONV.",
                             dger.convergencia_minimo_zsup)
            # Desconsidera vazão mínima
            escreve_alinhado("DESCONSIDERA VAZMIN",
                             dger.desconsidera_vazao_minima)
            # Considera restrições elétricas
            escreve_alinhado("RESTRICOES ELETRICAS",
                             dger.considera_restricoes_elet)
            # Seleção de cortes
            escreve_alinhado("SELECAO DE CORTES",
                             dger.selecao_cortes_benders)
            # Janela para seleção de cortes
            escreve_alinhado("JANELA DE CORTES",
                             dger.janela_selecao_cortes)
            # Reamostragem de cenários
            str_aux = "REAMOST. CENARIOS".ljust(len_aux)
            str_dados = ""
            d = int(dger.reamostragem)
            str_dados += str(int(d)).rjust(4) + " "
            d = int(dger.tipo_reamostragem.value)
            str_dados += str(int(d)).rjust(4) + " "
            d = int(dger.passo_reamostragem)
            str_dados += str(int(d)).rjust(4) + " "
            arq.write(str_aux + str_dados + "\n")
            # Considera nó 0 na convergência
            escreve_alinhado("CONVERGE NO ZERO",
                             dger.considera_convergencia_no0)
            # Consulta FCF
            escreve_alinhado("CONSULTA FCF",
                             dger.consulta_fcf)
            # Impressão ENA
            escreve_alinhado("IMPRESSAO ENA",
                             dger.impressao_ena)
            # Impressão de cortes ativos na simulação final
            escreve_alinhado("IMP. CATIVO S.FINAL",
                             dger.impressao_cortes_ativos)
            # Representante da agregação
            escreve_alinhado("REP. AGREGACAO",
                             dger.representante_agregacao.value)
            # Matriz de correlação espacial
            escreve_alinhado("MATRIZ CORR.ESPACIAL",
                             dger.matriz_corr_espacial.value)
            # Desconsidera convergência estatística
            escreve_alinhado("DESCONS. CONV. ESTAT",
                             dger.desconsidera_converg_estatist)
            # Momento da reamostragem
            escreve_alinhado("MOMENTO REAMOSTRAGEM",
                             dger.momento_reamostragem.value)
            # Mantém arquivos ENA
            escreve_alinhado("ARQUIVOS ENA",
                             dger.mantem_arquivos_ena)
            # Início do teste de convergência
            escreve_alinhado("INICIO TESTE CONVERG.",
                             dger.inicio_teste_convergencia.value)
            # Sazonalização Vmint
            escreve_alinhado("SAZ. VMINT PER. EST.",
                             dger.sazonaliza_vmint.value)
            # Sazonalização Vmaxt
            escreve_alinhado("SAZ. VMAXT PER. EST.",
                             dger.sazonaliza_vmaxt.value)
            # Sazonalização Vminp
            escreve_alinhado("SAZ. VMINP PER. EST.",
                             dger.sazonaliza_vminp.value)
            # Sazonalização Cfuga e Cmont
            escreve_alinhado("SAZ. CFUGA e CMONT",
                             dger.sazonaliza_cfuga_cmont.value)
            # Restrições de GEE
            escreve_alinhado("REST. EMISSAO GEE",
                             dger.restricoes_gee)
            # Uso do PAR(p)-A
            str_aux = "AFLUENCIA ANUAL PARP".ljust(len_aux)
            str_dados = ""
            d = int(dger.afluencia_anual_parp[0])
            str_dados += str(int(d)).rjust(4) + " "
            d = int(dger.afluencia_anual_parp[1])
            str_dados += str(int(not d)).rjust(4)
            arq.write(str_aux + str_dados + "\n")
            # Restrições de fornecimento de gás
            escreve_alinhado("REST. FORNEC. GAS",
                             dger.restricoes_fornecimento_gas)
            # Incerteza na geração eólica
            escreve_alinhado("INCERTEZA GER.EOLICA",
                             dger.incerteza_ger_eolica)
            # Incerteza na geração solar
            escreve_alinhado("INCERTEZA GER.SOLAR",
                             dger.incerteza_ger_solar)
            # Representação das incertezas
            escreve_alinhado("REPRESENTACAO INCERT",
                             dger.representacao_incerteza.value)



from typing import Dict
import numpy as np  # type: ignore
from inewave.config import NUM_CENARIOS, MESES


class Earmfpm00:
    """
    Armazena os dados das saídas referentes às energias
    armazenadas finais, por submercado e em % da energia armazenável máxima.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `earmfpm00x.out`, onde x varia conforme o
    submercado em questão.

    **Parâmetros**

    - mes_pmo: `int`
    - ano_pmo: `int`
    - versao_newave: `str`
    - submercado: `str`
    - energias_armazenadas: `Dict[int, np.ndarray]`

    """
    def __init__(self,
                 mes_pmo: int,
                 ano_pmo: int,
                 versao_newave: str,
                 submercado: str,
                 energias_armazenadas: Dict[int,
                                            np.ndarray]):
        self.mes_pmo = mes_pmo
        self.ano_pmo = ano_pmo
        self.versao_newave = versao_newave
        self.submercado = submercado
        self.energias_armazenadas = energias_armazenadas

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre Earmfpm00 avalia todos os valores, exceto
        a versão do NEWAVE.
        """
        if not isinstance(o, Earmfpm00):
            return False
        eaf: Earmfpm00 = o
        eq_mes_pmo = self.mes_pmo == eaf.mes_pmo
        eq_ano_pmo = self.ano_pmo == eaf.ano_pmo
        eq_submercado = self.submercado == eaf.submercado
        eq_e = all([np.array_equal(e, f)
                    for (e, f) in zip(self.energias_armazenadas.values(),
                                      eaf.energias_armazenadas.values())
                    ])
        return eq_mes_pmo and eq_ano_pmo and eq_submercado and eq_e

    @property
    def energias_por_ano(self) -> Dict[int, np.ndarray]:
        """
        Energias armazenadas para cada ano e em cada cenário, para
        todos os meses, organizadas primeiramente por ano.

        **Retorna**

        `Dict[int, np.ndarray]`

        **Sobre**

        O acesso é feito com [ano] e retorna um np.ndarray com os valores
        de EARM para todos os cenários e meses, naquele ano.

        """
        return self.energias_armazenadas

    @property
    def energias_por_ano_e_mes(self) -> Dict[int,
                                             Dict[int, np.ndarray]]:
        """
        Energias armazenadas para cada ano e mês em cada cenário,
        para todos os meses, organizadas primeiramente por ano.

        **Retorna**

        `Dict[int, Dict[int, np.ndarray]]`

        **Sobre**

        O acesso é feito com [ano][mes] e retorna um np.ndarray com os
        valores de EARM para todos os cenários, naquele ano e mês.

        """
        energias: Dict[int,
                       Dict[int, np.ndarray]] = {}
        anos_estudo = list(self.energias_armazenadas.keys())
        n_meses = len(MESES)
        # Cria e inicializa os objetos a serem retornados
        for a in anos_estudo:
            energias[a] = {m: np.array([])
                           for m in range(1, n_meses + 1)}
        # Preenche com os valores
        for a, tabela in self.energias_armazenadas.items():
            for m in range(1, n_meses + 1):
                col = m - 1
                energias[a][m] = tabela[:, col]
        return energias

    @property
    def energias_por_ano_e_cenario(self) -> Dict[int,
                                                 Dict[int, np.ndarray]]:
        """
        Energias armazenadas para cada ano e cenário, para todos os
        meses, organizadas primeiramente por ano.

        **Retorna**

        `Dict[int, Dict[int, np.ndarray]]`

        **Sobre**

        O acesso é feito com [ano][cenario] e retorna um np.ndarray com os
        valores de EARM para todos os meses, naquele ano e cenário.

        """
        energias: Dict[int,
                       Dict[int, np.ndarray]] = {}
        anos_estudo = list(self.energias_armazenadas.keys())
        # Cria e inicializa os objetos a serem retornados
        for a in anos_estudo:
            energias[a] = {c: np.array([])
                           for c in range(NUM_CENARIOS)}
        # Preenche com os valores
        for a, tabela in self.energias_armazenadas.items():
            for c in range(NUM_CENARIOS):
                energias[a][c] = tabela[c, :]
        return energias



# Rotinas de testes associadas ao arquivo parp.dat do NEWAVE
from inewave.newave.parp import LeituraPARp
from inewave.config import MESES, ORDEM_MAX_PARP
import numpy as np  # type: ignore

leitor = LeituraPARp("tests/_arquivos")
leitor.le_arquivo()

n_meses = len(MESES)


def test_leitura():
    assert leitor.parp.series_energia[1].shape == (2018 - 1931 + 1,
                                                   n_meses + 1,
                                                   50)
    assert leitor.parp.ordens_orig[1].shape == (5, n_meses + 1)
    assert leitor.parp.coeficientes[1].shape == (5 * n_meses,
                                                 ORDEM_MAX_PARP,
                                                 4)


def test_eq_parp():
    leitor2 = LeituraPARp("tests/_arquivos")
    leitor2.le_arquivo()
    assert leitor.parp == leitor2.parp


def test_neq_parp():
    leitor2 = LeituraPARp("tests/_arquivos")
    leitor2.le_arquivo()
    leitor2.parp.ordens_orig[1] = np.array([])
    assert leitor2.parp != leitor.parp
    assert leitor2.parp is not None


def test_series_energia_ree():
    series = leitor.parp.series_energia_ree(1)
    assert len(series.keys()) == 50


def test_series_medias_ree():
    series = leitor.parp.series_medias_ree(1)
    assert len(series.keys()) == 5


def test_correlograma_energia_ree():
    correl = leitor.parp.correlograma_energia_ree(1)
    assert len(correl.keys()) == 5 * n_meses


def test_correlograma_media_ree():
    correl = leitor.parp.correlograma_media_ree(1)
    assert len(correl.keys()) == 5 * n_meses


def test_ordens_originais_ree():
    ordens = leitor.parp.ordens_originais_ree(1)
    assert len(ordens.keys()) == 5


def test_ordens_finais_ree():
    ordens = leitor.parp.ordens_finais_ree(1)
    assert len(ordens.keys()) == 5


def test_coeficientes_ree():
    coefs = leitor.parp.coeficientes_ree(1)
    assert len(coefs) == 5 * n_meses


def test_correlacoes_espaciais_ano_configuracao():
    corrs = leitor.parp.correlacoes_espaciais_anuais
    assert len(corrs.keys()) == 3


def test_correlacoes_espaciais_mes_configuracao():
    corrs = leitor.parp.correlacoes_espaciais_mensais
    assert len(corrs.keys()) == 3



# Rotinas de testes associadas ao arquivo MEDIAS-SIN.CSV do NWLISTOP
from inewave.nwlistop.mediassin import LeituraMediasSIN
import numpy as np  # type: ignore


leitor = LeituraMediasSIN("tests/_arquivos")
leitor.le_arquivo()


def test_leitura():
    assert leitor.medias.mes_pmo != 0


def test_eq_mediasmerc():
    leitor2 = LeituraMediasSIN("tests/_arquivos")
    leitor2.le_arquivo()
    assert leitor.medias == leitor2.medias


def test_neq_mediassin():
    leitor2 = LeituraMediasSIN("tests/_arquivos")
    leitor2.le_arquivo()
    leitor2.medias.tabela = np.array([])
    assert leitor.medias != leitor2.medias


def test_alinhamento_vertical():
    mes_pmo = leitor.medias.mes_pmo
    # Confere se todos os valores anteriores ao mês
    # do PMO são nulos
    assert np.all(leitor.medias.tabela[:, :mes_pmo-1] == 0)


def test_alinhamento_horizontal():
    # Confere se todos os valores de geracao eolica e solar
    # são nulos
    eol = leitor.medias.geracao_eolica
    sol = leitor.medias.geracao_solar
    for e, s in zip(eol, sol):
        assert np.all(e == 0.0)
        assert np.all(s == 0.0)


def test_valores_vertimento():
    # Confere se todos os valores de vertimentos controláveis
    # e fio d'água somam o total, a menos da precisão.
    vert_total = leitor.medias.vertimento_total
    vert_control = leitor.medias.vertimento_controlavel
    vert_fio = leitor.medias.vertimento_fio_dagua
    for t, c, f in zip(vert_total,
                       vert_control,
                       vert_fio):
        assert np.all(t - c - f < 1e-1)


def test_energias_armazenadas():
    e = leitor.medias.energias_armazenadas_absolutas
    assert len(e) > 0
    e = leitor.medias.energias_armazenadas_percentuais
    assert len(e) > 0


def test_percentis_energias():
    e = leitor.medias.percentil_10_energias_armazenadas
    assert len(e) > 0
    e = leitor.medias.percentil_20_energias_armazenadas
    assert len(e) > 0
    e = leitor.medias.percentil_30_energias_armazenadas
    assert len(e) > 0
    e = leitor.medias.percentil_40_energias_armazenadas
    assert len(e) > 0
    e = leitor.medias.percentil_50_energias_armazenadas
    assert len(e) > 0
    e = leitor.medias.percentil_60_energias_armazenadas
    assert len(e) > 0
    e = leitor.medias.percentil_70_energias_armazenadas
    assert len(e) > 0
    e = leitor.medias.percentil_80_energias_armazenadas
    assert len(e) > 0
    e = leitor.medias.percentil_90_energias_armazenadas
    assert len(e) > 0
    e = leitor.medias.percentil_100_energias_armazenadas
    assert len(e) > 0


def test_energias_afluentes():
    e = leitor.medias.energia_natural_afluente
    assert len(e) > 0
    e = leitor.medias.energia_controlavel_corrigida
    assert len(e) > 0
    e = leitor.medias.energia_fio_dagua_bruta
    assert len(e) > 0
    e = leitor.medias.energia_fio_dagua_liquida
    assert len(e) > 0
    e = leitor.medias.energia_evaporada
    assert len(e) > 0


def test_geracoes():
    e = leitor.medias.geracao_hidraulica_controlavel
    assert len(e) > 0
    e = leitor.medias.geracao_hidraulica_fio_dagua_liquida
    assert len(e) > 0
    e = leitor.medias.geracao_hidraulica_total
    assert len(e) > 0
    e = leitor.medias.geracao_eolica
    assert len(e) > 0
    e = leitor.medias.geracao_solar
    assert len(e) > 0
    e = leitor.medias.geracao_termica
    assert len(e) > 0


def test_custos():
    e = leitor.medias.deficit
    assert len(e) > 0
    e = leitor.medias.custo_deficit
    assert len(e) > 0
    e = leitor.medias.custo_termica
    assert len(e) > 0



from typing import Dict
import numpy as np  # type: ignore
from inewave.newave.modelos.patamar import Patamar
from inewave.config import NUM_PATAMARES, NUM_CENARIOS, MESES


class Cmarg00:
    """
    Armazena os dados das saídas referentes aos custos marginais de operação
    por patamar, por submercado.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `cmarg00x.out`, onde x varia conforme o
    submercado em questão.

    **Parâmetros**

    - mes_pmo: `int`
    - ano_pmo: `int`
    - versao_newave: `str`
    - submercado: `str`
    - custos_patamares: `Dict[int, np.ndarray]`

    """
    def __init__(self,
                 mes_pmo: int,
                 ano_pmo: int,
                 versao_newave: str,
                 submercado: str,
                 custos_patamares: Dict[int,
                                        np.ndarray]):
        self.mes_pmo = mes_pmo
        self.ano_pmo = ano_pmo
        self.versao_newave = versao_newave
        self.submercado = submercado
        self.custos_patamares = custos_patamares

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre Cmarg00 avalia todos os valores, exceto
        a versão do NEWAVE.
        """
        if not isinstance(o, Cmarg00):
            return False
        cmarg: Cmarg00 = o
        eq_mes_pmo = self.mes_pmo == cmarg.mes_pmo
        eq_ano_pmo = self.ano_pmo == cmarg.ano_pmo
        eq_submercado = self.submercado == cmarg.submercado
        eq_custos = all([np.array_equal(c1, c2)
                         for (c1, c2) in zip(self.custos_patamares.values(),
                                             cmarg.custos_patamares.values())
                         ])
        return eq_mes_pmo and eq_ano_pmo and eq_submercado and eq_custos

    def custos_medios_por_ano(self,
                              patamar: Patamar) -> Dict[int,
                                                        np.ndarray]:
        """
        Custos médios para cada ano de estudo.

        **Parâmetros**

        `Patamar`

        **Retorna**

        `Dict[int, np.ndarray]`

        **Sobre**

        Recebe um objeto Patamar para realizar a ponderação de
        cada cenários pelos devidos valores. O acesso é feito com
        [ano] e a saída é uma array 2-D do numpy com os valores
        médios de cada cenário e mês.
        """
        patamares_anos = patamar.patamares_por_ano
        # Confere se os anos de estudo do objeto Patamar fornecido
        # são os mesmos do cmarg00
        if patamar.anos_estudo != list(self.custos_patamares.keys()):
            raise Exception("Objeto Patamar incompatível com Cmarg00")
        # Inicializa a variável que irá conter os custos médios
        n_meses = len(MESES)
        custos: Dict[int, np.ndarray] = {a: np.zeros((NUM_CENARIOS,
                                                      n_meses))
                                         for a in patamar.anos_estudo}
        # Para cada cenário, calcula o custo médio em relação aos
        # patamares
        for a in patamar.anos_estudo:
            for i in range(NUM_CENARIOS):
                li = i * NUM_PATAMARES
                lf = li + NUM_PATAMARES
                janela: np.ndarray = self.custos_patamares[a][li:lf, :]
                custos[a][i, :] = np.inner(janela.T,
                                           patamares_anos[a].T)[:, 0]
        return custos

    def custos_medios_por_ano_e_mes(self,
                                    p: Patamar) -> Dict[int,
                                                        Dict[int,
                                                             np.ndarray]]:
        """
        Custos médios para cada ano de estudo e mês.

        **Parâmetros**

        `Patamar`

        **Retorna**

        `Dict[int, Dict[int, np.ndarray]]`

        **Sobre**

        Recebe um objeto Patamar para realizar a ponderação de
        cada cenários pelos devidos valores. O acesso é feito com
        [ano][mes] e a saída é uma array do numpy com os valores
        médios de cada cenário.
        """
        custos_ano = self.custos_medios_por_ano(p)
        n_meses = len(MESES)
        # Inicializa a variável que irá conter os custos médios por mês
        custos: Dict[int,
                     Dict[int,
                          np.ndarray]] = {a: {}
                                          for a in p.anos_estudo}
        # Separa os custos em cada mês e retorna
        for a in p.anos_estudo:
            for m in range(1, n_meses + 1):
                custos[a][m] = custos_ano[a][:, m - 1]
        return custos

    @property
    def custos_por_patamar(self) -> Dict[int,
                                         Dict[int,
                                              Dict[int, np.ndarray]]]:
        """
        Custos obtidos para cada ano, mês e em cada cenário,
        organizados primeiramente por patamar.

        **Retorna**

        `Dict[int, Dict[int, Dict[int, np.ndarray]]]`

        **Sobre**

        O acesso é feito
        com [patamar][ano][mes] e retorna um np.ndarray.
        """
        custos: Dict[int,
                     Dict[int,
                          Dict[int, np.ndarray]]] = {}
        patamares = list(range(1, NUM_PATAMARES + 1))
        anos_estudo = list(self.custos_patamares.keys())
        n_meses = len(MESES)
        # Cria e inicializa os objetos a serem retornados
        for p in patamares:
            if p not in custos:
                custos[p] = {a: {m: np.zeros(NUM_CENARIOS,)
                                 for m in range(1, n_meses + 1)}
                             for a in anos_estudo}
        # Preenche com os valores
        for a, tabela in self.custos_patamares.items():
            for p in patamares:
                for c in range(NUM_CENARIOS):
                    for m in range(1, n_meses + 1):
                        lin = NUM_PATAMARES * c + p - 1
                        col = m - 1
                        custos[p][a][m][c] = tabela[lin, col]
        return custos

    @property
    def custos_por_ano(self) -> Dict[int,
                                     Dict[int, np.ndarray]]:
        """
        Custos obtidos para cada ano e em cada cenário, para
        todos os meses, organizados primeiramente por ano.

        **Retorna**

        `Dict[int, Dict[int,  np.ndarray]]`

        **Sobre**

        O acesso é feito
        com [patamar][ano] e retorna um np.ndarray com os valores de
        custos para todos os cenários e meses, naquele patamar e ano.
        """
        custos: Dict[int,
                     Dict[int, np.ndarray]] = {}
        patamares = list(range(1, NUM_PATAMARES + 1))
        anos_estudo = list(self.custos_patamares.keys())
        n_meses = len(MESES)
        # Cria e inicializa os objetos a serem retornados
        for p in patamares:
            if p not in custos:
                custos[p] = {a:  np.zeros((NUM_CENARIOS, n_meses))
                             for a in anos_estudo}
        # Preenche com os valores
        for a, tabela in self.custos_patamares.items():
            for p in patamares:
                for c in range(NUM_CENARIOS):
                    lin = NUM_PATAMARES * c + p - 1
                    custos[p][a][c, :] = tabela[lin, :]
        return custos

    @property
    def custos_por_ano_e_mes(self) -> Dict[int,
                                           Dict[int,
                                                Dict[int, np.ndarray]]]:
        """
        Custos obtidos para cada ano, mês e em cada cenário,
        organizados primeiramente por ano e mês.

        **Retorna**

        `Dict[int, Dict[int, Dict[int, np.ndarray]]]`

        **Sobre**

        O acesso é feito com
        [patamar][ano][mes] e retorna um np.ndarray com os valores de
        custos para todos os cenários, naquele patamar.
        """
        custos: Dict[int,
                     Dict[int,
                          Dict[int, np.ndarray]]] = {}
        patamares = list(range(1, NUM_PATAMARES + 1))
        anos_estudo = list(self.custos_patamares.keys())
        n_meses = len(MESES)
        # Cria e inicializa os objetos a serem retornados
        for p in patamares:
            if p not in custos:
                custos[p] = {a: {m: np.zeros(NUM_CENARIOS,)
                                 for m in range(1, n_meses + 1)}
                             for a in anos_estudo}
        # Preenche com os valores
        for a, tabela in self.custos_patamares.items():
            for p in patamares:
                for c in range(NUM_CENARIOS):
                    for m in range(1, n_meses + 1):
                        lin = NUM_PATAMARES * c + p - 1
                        col = m - 1
                        custos[p][a][m][c] = tabela[lin, col]
        return custos



from typing import List, Dict
import numpy as np  # type: ignore
from inewave.config import MESES


class Patamar:
    """
    Armazena os dados de entrada do NEWAVE referentes aos
    patamares de carga por submercado.

    Esta classe pode lidar com um número qualquer de patamares
    de carga, desde que as informações fornecidas a ela por meio
    da tabela de valores seja compatível com o parâmetro `num_patamares`
    da mesma.

    A tabela de patamares de carga é armazenada através de uma array
    em `NumPy`, para otimizar cálculos futuros e espaço ocupado
    em memória. A tabela interna é transformada em dicionários
    e outras estruturas de dados mais palpáveis através das propriedades
    da própria classe.

    **Parâmetros**

    - num_patamares: `int`
    - anos_estudo: `List[int]`
    - paramares: `np.ndarray`

    """
    def __init__(self,
                 num_patamares: int,
                 anos_estudo: List[int],
                 patamares: np.ndarray):
        self.num_patamares = num_patamares
        self.anos_estudo = anos_estudo
        self.patamares = patamares

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre Patamar avalia todos os campos.
        """
        if not isinstance(o, Patamar):
            return False
        pat: Patamar = o
        eq_num = self.num_patamares == pat.num_patamares
        eq_anos = self.anos_estudo == pat.anos_estudo
        eq_pats = np.array_equal(self.patamares, pat.patamares)
        return eq_num and eq_anos and eq_pats

    @property
    def patamares_por_ano(self) -> Dict[int, np.ndarray]:
        """
        Valores contidos na tabela de patamares, organizados
        por ano.

        **Retorna**

        `Dict[int, np.ndarray]`

        **Sobre**

        O acesso é feito com [ano] e o valor fornecido
        é uma array 2-D do `NumPy` com os valores dos patamares para todo
        os meses de um ano, semelhante a uma linha do arquivo patamar.dat.
        """
        patamares_ano: Dict[int, np.ndarray] = {}
        # Preenche com os valores
        for i, a in enumerate(self.anos_estudo):
            li = self.num_patamares * i
            lf = li + self.num_patamares
            patamares_ano[a] = self.patamares[li:lf, :]
        return patamares_ano

    @property
    def patamares_por_ano_e_mes(self) -> Dict[int, Dict[int, np.ndarray]]:
        """
        Valores contidos na tabela de patamares, organizados por ano e mês.

        **Retorna**

        `Dict[int, Dict[int, np.ndarray]]`

        **Sobre**

        O acesso é feito com [ano][mes] e o valor fornecido
        é uma array do numpy com os valores para os patamares 1, 2, ... .
        """
        patamares_ano: Dict[int, Dict[int, np.ndarray]] = {}
        n_meses = len(MESES)
        # Cria e inicializa os objetos
        for a in self.anos_estudo:
            patamares_ano[a] = {m: np.zeros(self.num_patamares,)
                                for m in range(1, n_meses + 1)}
        # Preenche com os valores
        for i, a in enumerate(self.anos_estudo):
            for m in range(1, n_meses + 1):
                li = self.num_patamares * i
                lf = li + self.num_patamares
                col = m - 1
                patamares_ano[a][m] = self.patamares[li:lf, col]
        return patamares_ano



from typing import Dict
import numpy as np  # type: ignore
from inewave.config import SUBMERCADOS, NUM_VARIAVEIS_MEDIAS


class MediasMerc:
    """
    Armazena os dados das saídas referentes às médias de diversas variáveis
    agrupadas por submercado.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `MEDIAS-MERC.CSV`.

    **Parâmetros**

    - mes_pmo: `int`
    - tabela: `np.ndarray`

    """
    def __init__(self,
                 mes_pmo: int,
                 tabela: np.ndarray):
        self.mes_pmo = mes_pmo
        self.tabela = tabela

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre MediasMerc avalia todos os valores da tabela.
        """
        if not isinstance(o, MediasMerc):
            return False
        medias: MediasMerc = o
        eq_mes_pmo = self.mes_pmo == medias.mes_pmo
        return eq_mes_pmo and np.array_equal(self.tabela, medias.tabela)

    def _extrai_variavel_tabela(self,
                                indice_variavel: int) -> Dict[str,
                                                              np.ndarray]:
        """
        Lógica para extrair uma variável qualquer da tabela de médias,
        partindo do índice dela (num. da linha) dentro do seu submercado.
        """
        valores: Dict[str, np.ndarray] = {sub: np.array([])
                                          for sub in SUBMERCADOS}
        for i, sub in enumerate(SUBMERCADOS):
            lin = NUM_VARIAVEIS_MEDIAS * i + indice_variavel
            valores[sub] = self.tabela[lin, :]
        return valores

    @property
    def energias_armazenadas_absolutas(self) -> Dict[str,
                                                     np.ndarray]:
        """
        Energias armazenadas em valores absolutos (MWmed) para
        cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida
        mensalmente, para todo o período de estudo.

        """
        return self._extrai_variavel_tabela(0)

    @property
    def energias_armazenadas_percentuais(self) -> Dict[str,
                                                       np.ndarray]:
        """
        Energias armazenadas em valores em percentual da EARMax
        para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida
        mensalmente, para todo o período de estudo.

        """
        return self._extrai_variavel_tabela(1)

    @property
    def percentil_10_energias_armazenadas(self) -> Dict[str,
                                                        np.ndarray]:
        """
        Valores para o percentil 10% da energia armazenada
        para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida
        mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(2)

    @property
    def percentil_20_energias_armazenadas(self) -> Dict[str,
                                                        np.ndarray]:
        """
        Valores para o percentil 20% da energia armazenada
        para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida
        mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(3)

    @property
    def percentil_30_energias_armazenadas(self) -> Dict[str,
                                                        np.ndarray]:
        """
        Valores para o percentil 30% da energia armazenada
        para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida
        mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(4)

    @property
    def percentil_40_energias_armazenadas(self) -> Dict[str,
                                                        np.ndarray]:
        """
        Valores para o percentil 40% da energia armazenada
        para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida
        mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(5)

    @property
    def percentil_50_energias_armazenadas(self) -> Dict[str,
                                                        np.ndarray]:
        """
        Valores para o percentil 50% da energia armazenada
        para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida
        mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(6)

    @property
    def percentil_60_energias_armazenadas(self) -> Dict[str,
                                                        np.ndarray]:
        """
        Valores para o percentil 60% da energia armazenada
        para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida
        mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(7)

    @property
    def percentil_70_energias_armazenadas(self) -> Dict[str,
                                                        np.ndarray]:
        """
        Valores para o percentil 70% da energia armazenada
        para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida
        mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(8)

    @property
    def percentil_80_energias_armazenadas(self) -> Dict[str,
                                                        np.ndarray]:
        """
        Valores para o percentil 80% da energia armazenada
        para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida
        mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(9)

    @property
    def percentil_90_energias_armazenadas(self) -> Dict[str,
                                                        np.ndarray]:
        """
        Valores para o percentil 90% da energia armazenada
        para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida
        mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(10)

    @property
    def percentil_100_energias_armazenadas(self) -> Dict[str,
                                                         np.ndarray]:
        """
        Valores para o percentil 100% da energia armazenada
        para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida
        mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(11)

    @property
    def energia_natural_afluente(self) -> Dict[str,
                                               np.ndarray]:
        """
        Energias naturais afluentes em valores absolutos (MWmed)
        para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida
        mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(12)

    @property
    def energia_controlavel_corrigida(self) -> Dict[str,
                                                    np.ndarray]:
        """
        Energias controláveis corrigidas em valores absolutos
        (MWmed) para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada
        é dividida mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(13)

    @property
    def energia_fio_dagua_bruta(self) -> Dict[str,
                                              np.ndarray]:
        """
        Energias fio d'água brutas em valores absolutos
        (MWmed) para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada
        é dividida mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(14)

    @property
    def energia_fio_dagua_liquida(self) -> Dict[str,
                                                np.ndarray]:
        """
        Energias fio d'água líquidas em valores absolutos
        (MWmed) para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada
        é dividida mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(15)

    @property
    def geracao_hidraulica_maxima(self) -> Dict[str,
                                                np.ndarray]:
        """
        Gerações hidráulicas máximas em valores absolutos
        (MWmed) para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada
        é dividida mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(16)

    @property
    def geracao_hidraulica_total(self) -> Dict[str,
                                               np.ndarray]:
        """
        Gerações hidráulicas totais em valores absolutos
        (MWmed) para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada
        é dividida mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(18)

    @property
    def geracao_hidraulica_controlavel(self) -> Dict[str,
                                                     np.ndarray]:
        """
        Gerações hidráulicas controláveis em valores absolutos
        (MWmed) para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada
        é dividida mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(19)

    @property
    def geracao_hidraulica_fio_dagua_liquida(self) -> Dict[str,
                                                           np.ndarray]:
        """
        Gerações hidráulicas fio d'água líquidas em valores
        absolutos (MWmed) para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica
        retornada é dividida mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(20)

    @property
    def geracao_eolica(self) -> Dict[str,
                                     np.ndarray]:
        """
        Gerações eólicas em valores absolutos (MWmed) para cada um
        dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida mensalmente,
        para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(21)

    @property
    def geracao_solar(self) -> Dict[str,
                                    np.ndarray]:
        """
        Gerações solares em valores absolutos (MWmed) para cada um
        dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida mensalmente,
        para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(22)

    @property
    def vertimento_total(self) -> Dict[str,
                                       np.ndarray]:
        """
        Perdas por vertimento totais em valores absolutos (MWmed)
        para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida
        mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(23)

    @property
    def vertimento_controlavel(self) -> Dict[str,
                                             np.ndarray]:
        """
        Perdas por vertimento controlável em valores absolutos
        (MWmed) para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada
        é dividida mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(24)

    @property
    def vertimento_fio_dagua(self) -> Dict[str,
                                           np.ndarray]:
        """
        Perdas por vertimento fio d'água em valores absolutos
        (MWmed) para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada
        é dividida mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(25)

    @property
    def vertimento_turbinavel(self) -> Dict[str,
                                            np.ndarray]:
        """
        Vertimentos turbináveis em valores absolutos
        (MWmed) para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada
        é dividida mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(26)

    @property
    def energia_evaporada(self) -> Dict[str,
                                        np.ndarray]:
        """
        Perdas por energia evaporada em valores absolutos
        (MWmed) para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada
        é dividida mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(40)

    @property
    def geracao_termica(self) -> Dict[str,
                                      np.ndarray]:
        """
        Gerações de usinas termelétricas em valores absolutos
        (MWmed) para cada um dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada
        é dividida mensalmente, para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(42)

    @property
    def deficit(self) -> Dict[str,
                              np.ndarray]:
        """
        Déficits de carga em valores absolutos (MWmed) para cada um
        dos submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida mensalmente,
        para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(43)

    @property
    def custo_marginal_operacao(self) -> Dict[str,
                                              np.ndarray]:
        """
        Custos marginais de operação (R$) para cada um dos
        submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida mensalmente,
        para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(45)

    @property
    def custo_deficit(self) -> Dict[str,
                                    np.ndarray]:
        """
        Custos de déficit (R$) para cada um dos
        submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida mensalmente,
        para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(46)

    @property
    def custo_termica(self) -> Dict[str,
                                    np.ndarray]:
        """
        Custos de térmicas (R$) para cada um dos
        submercados.

        **Retorna**

        `Dict[str, np.ndarray]`

        **Sobre**

        A série histórica retornada é dividida mensalmente,
        para todo o período de estudo.
        """
        return self._extrai_variavel_tabela(47)



# Rotinas de testes associadas ao arquivo eafbm00x.out do NWLISTOP
from inewave.nwlistop.eafbm00 import LeituraEafbm00
from inewave.config import MESES, NUM_CENARIOS
import numpy as np  # type: ignore


sub_teste = "SUDESTE"
leitor = LeituraEafbm00("tests/_arquivos")
leitor.le_arquivos()
anos_estudo_teste = [1995, 1996, 1997, 1998, 1999]


def test_leitura():
    assert sub_teste in leitor.eafbms


def test_eq_eafbm00():
    leitor2 = LeituraEafbm00("tests/_arquivos")
    leitor2.le_arquivos()
    assert leitor.eafbms == leitor2.eafbms


def test_neq_eafbm00():
    leitor2 = LeituraEafbm00("tests/_arquivos")
    leitor2.le_arquivos()
    leitor2.eafbms[sub_teste].ano_pmo = 2077
    assert leitor.eafbms != leitor2.eafbms


def test_extrai_dados_execucao():
    eafbm = leitor.eafbms[sub_teste]
    assert eafbm.mes_pmo == 1
    assert eafbm.ano_pmo == 1995
    assert eafbm.submercado == sub_teste


def test_anos_estudo():
    eafbm = leitor.eafbms[sub_teste]
    anos_lidos = list(eafbm.energias_afluentes.keys())
    assert anos_estudo_teste == anos_lidos


def test_eafbm_por_ano():
    eafbm = leitor.eafbms[sub_teste]
    por_ano = eafbm.energias_por_ano
    assert anos_estudo_teste == list(por_ano.keys())
    # Confere se os valores de energia afluente são sempre não nulos
    for a in anos_estudo_teste:
        assert np.all(por_ano[a] > 0.0)


def test_eafbm_por_ano_e_mes():
    eafbm = leitor.eafbms[sub_teste]
    por_ano_e_mes = eafbm.energias_por_ano_e_mes
    assert anos_estudo_teste == list(por_ano_e_mes.keys())
    # Confere se os valores de energia afluente são sempre não nulos
    for a in anos_estudo_teste:
        for m in range(1, len(MESES) + 1):
            assert np.all(por_ano_e_mes[a][m] > 0.0)


def test_eafbm_por_cenario():
    eafbm = leitor.eafbms[sub_teste]
    por_ano_cenario = eafbm.energias_por_ano_e_cenario
    # Confere se os valores de energia afluente são sempre não nulos
    n_meses = len(MESES)
    for a in anos_estudo_teste:
        for c in range(NUM_CENARIOS):
            assert por_ano_cenario[a][c].shape == (n_meses,)
            assert np.all(por_ano_cenario[a][c] > 0)



# Rotinas de testes associadas ao arquivo patamar.dat do NEWAVE
from inewave.newave.patamar import LeituraPatamar
from inewave.config import MESES
import numpy as np  # type: ignore


anos_estudo_teste = [1995, 1996, 1997, 1998, 1999]
leitor = LeituraPatamar("tests/_arquivos")
leitor.le_arquivo()


def test_leitura():
    assert leitor.patamar.num_patamares != 0


def test_eq_patamar():
    leitor2 = LeituraPatamar("tests/_arquivos")
    leitor2.le_arquivo()
    assert leitor.patamar == leitor2.patamar


def test_neq_patamar():
    leitor2 = LeituraPatamar("tests/_arquivos")
    leitor2.le_arquivo()
    leitor2.patamar.anos_estudo = []
    assert leitor.patamar != leitor2.patamar


def test_anos_estudo():
    assert anos_estudo_teste == leitor.patamar.anos_estudo


def test_leitura_tabela():
    assert np.all(leitor.patamar.patamares >= 0.0)


def test_patamar_por_ano():
    patamares_ano = leitor.patamar.patamares_por_ano
    assert anos_estudo_teste == list(patamares_ano.keys())
    # Verifica as dimensões de cada conjunto de patamares
    # e se todos os valores são maiores que 0
    n_meses = len(MESES)
    for ano in anos_estudo_teste:
        p = patamares_ano[ano]
        assert p.shape == (leitor.patamar.num_patamares,
                           n_meses)
        assert np.all(p > 0.0)


def test_patamar_por_ano_e_mes():
    patamares_ano = leitor.patamar.patamares_por_ano_e_mes
    assert anos_estudo_teste == list(patamares_ano.keys())
    # Verifica as dimensões de cada conjunto de patamares
    # e se todos os valores são maiores que 0
    for ano in anos_estudo_teste:
        for mes in range(1, len(MESES) + 1):
            p = patamares_ano[ano][mes]
            assert p.shape == (leitor.patamar.num_patamares,)
            assert np.all(p > 0.0)



# Rotinas de testes associadas ao arquivo confhd.dat do NEWAVE
from inewave.newave.confhd import LeituraConfhd, EscritaConfhd


leitor = LeituraConfhd("tests/_arquivos")
leitor.le_arquivo()


def test_leitura():
    assert len(leitor.confhd.usinas) > 0


def test_escrita_e_leitura():
    escritor = EscritaConfhd("tests/_saidas")
    escritor.escreve_arquivo(leitor.confhd)
    leitor2 = LeituraConfhd("tests/_saidas")
    leitor2.le_arquivo()
    assert leitor.confhd == leitor2.confhd


def test_eq_confhd():
    leitor2 = LeituraConfhd("tests/_arquivos")
    leitor2.le_arquivo()
    assert leitor2.confhd == leitor.confhd


def test_neq_confhd():
    leitor2 = LeituraConfhd("tests/_arquivos")
    leitor2.le_arquivo()
    leitor2.confhd.usinas[4].existente = False
    assert leitor2.confhd != leitor.confhd
    assert leitor2.confhd is not None



# Imports do próprio módulo
from inewave.nwlistop.modelos.mediasmerc import MediasMerc
from inewave._utils.leitura import Leitura
from inewave.config import MAX_ANOS_ESTUDO, MESES
from inewave.config import NUM_VARIAVEIS_MEDIAS, SUBMERCADOS
# Imports de módulos externos
import os
import csv
import numpy as np  # type: ignore
from traceback import print_exc


class LeituraMediasMerc(Leitura):
    """
    Realiza a leitura do arquivo MEDIAS-MERC.CSV
    existente em um diretório de saídas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos do arquivo MEDIAS-MERC.CSV, construindo um
    objeto `MediasMerc` cujas informações são as mesmas do arquivo.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.

    Uma vez realizada a leitura do arquivo, as informações são guardadas
    internamente no atributo `medias`.

    **Exemplos**

    >>> diretorio = "~/documentos/.../deck"
    >>> leitor = LeituraMediasMerc(diretorio)
    >>> leitor.le_arquivo()
    # Ops, esqueci de pegar o objeto
    >>> medias = leitor.medias

    """
    def __init__(self,
                 diretorio: str) -> None:
        super().__init__()
        self.diretorio = diretorio
        # MediasMerc default, depois é substituído
        self.medias = MediasMerc(0, np.array([]))

    def le_arquivo(self) -> MediasMerc:
        """
        Lê um arquivo MEDIAS-MERC.CSV em um diretório.
        """
        try:
            caminho = os.path.join(self.diretorio, "MEDIAS-MERC.CSV")
            n_meses = len(MESES)
            n_submercados = len(SUBMERCADOS)
            linhas_medias = NUM_VARIAVEIS_MEDIAS * n_submercados
            colunas_medias = MAX_ANOS_ESTUDO * n_meses
            mes_pmo = 0
            n_colunas = 0
            with open(caminho, newline='') as arq:
                leitor = csv.reader(arq, delimiter=",", quotechar='|')
                tabela = np.zeros((linhas_medias, colunas_medias))
                # Descobre o mês do PMO pela terceira coluna
                # da primeira linha. Depois lê linha a linha.
                primeira_linha = True
                for i, linha in enumerate(leitor):
                    if primeira_linha:
                        mes_pmo = int(linha[2])
                        n_colunas = len(linha) + mes_pmo - 4
                        primeira_linha = False
                        continue
                    # Ignora as linhas após os dados dos submercados
                    if i > linhas_medias:
                        break
                    linha_num = np.array([float(n) for n in linha[2:-1]])
                    tabela[i - 1, mes_pmo-1:n_colunas] = linha_num
                self.medias = MediasMerc(mes_pmo, tabela[:, :n_colunas])
                return self.medias
        except Exception:
            print_exc()
            return MediasMerc(0, np.array([]))

    def _fim_arquivo(self, linha: str) -> bool:
        return False



# Imports do próprio módulo
from inewave._utils.escrita import Escrita
from inewave._utils.leitura import Leitura
from .modelos.dsvagua import DSVAgua
from inewave.config import MESES
# Imports de módulos externos
import os
import numpy as np  # type: ignore
from traceback import print_exc
from typing import IO


class LeituraDSVAgua(Leitura):
    """
    Realiza a leitura do arquivo `dsvagua.dat`
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo `dsvagua.dat`, construindo
    um objeto `DSVAgua` cujas informações são as mesmas do `dsvagua.dat`.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.

    Uma vez realizada a leitura do arquivo, as informações são guardadas
    internamente no atributo `dsvagua`.

    **Exemplos**

    >>> diretorio = "~/documentos/.../deck"
    >>> leitor = LeituraDSVAgua(diretorio)
    >>> leitor.le_arquivo()
    # Ops, esqueci de pegar o objeto
    >>> dsv = leitor.dsvagua

    """

    def __init__(self,
                 diretorio: str) -> None:
        super().__init__()
        self.diretorio = diretorio
        # Confhd default, depois é substituído
        self.dsvagua = DSVAgua(np.array([]))

    def le_arquivo(self) -> DSVAgua:
        """
        Faz a leitura do arquivo `dsvagua.dat`.
        """
        try:
            caminho = os.path.join(self.diretorio, "dsvagua.dat")
            with open(caminho, "r") as arq:
                self.confhd = self._le_dsvagua(arq)
                return self.dsvagua
        except Exception:
            print_exc()
            return self.dsvagua

    def _le_dsvagua(self, arq: IO) -> DSVAgua:
        """
        Faz a leitura do arquivo dsvagua.dat.
        """
        linha = ""
        # Pula as duas primeiras linhas, com cabeçalhos
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        n_meses = len(MESES)
        tabela = np.zeros((5000, n_meses + 3))
        i = 0
        while True:
            # Verifica se o arquivo acabou
            linha = self._le_linha_com_backup(arq)
            if self._fim_arquivo(linha):
                self.dsvagua = DSVAgua(tabela[:i, :])
                break
            # Senão, lê mais uma linha
            # Ano
            tabela[i, 0] = int(linha[:4])
            # Usina
            tabela[i, 1] = int(linha[6:9])
            # Desvios de cada mês
            ci = 10
            nc = 6
            for m in range(n_meses):
                cf = ci + nc
                valor = float(linha[ci:cf])
                tabela[i, m + 2] = valor
                ci = cf + 1
            tabela[i, -1] = int(linha[97])
            i += 1

        return self.dsvagua

    def _fim_arquivo(self, linha: str) -> bool:
        return linha.strip() == "9999"


class EscritaDSVAgua(Escrita):
    """
    Realiza a escrita do arquivo dsvagua.dat
    em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para escrever os campos
    de um arquivo dsvagua.dat, a partir de um objeto `DSVAgua`.

    Este objeto existe para retirar do modelo de dados a complexidade
    de armazenar as strings auxiliares do arquivo, desenhar tabelas, dentre
    outras tarefas associadas à escrita.

    Se o diretório de escrita não existir, ele será criado.

    **Exemplos**

    >>> diretorio = "~/documentos/.../deck"
    >>> # dsv é do tipo DSVAgua
    >>> escritor = EscritaDSVAgua(diretorio)
    >>> escritor.escreve_arquivo(dsv)
    """
    def __init__(self, diretorio: str):
        super().__init__()
        self.diretorio = diretorio

    def escreve_arquivo(self, dsvagua: DSVAgua):
        """
        Realiza a escrita de um arquivo `dsvagua.dat`.
        """
        # Confere se o diretório existe. Senão, cria.
        if not os.path.exists(self.diretorio):
            os.makedirs(self.diretorio)
        # Inicia a escrita
        caminho = os.path.join(self.diretorio, "dsvagua.dat")
        n_meses = len(MESES)
        with open(caminho, "w") as arq:

            def escreve_desvios(dsv: DSVAgua):
                lin_tab = dsv.tabela.shape[0]
                for i in range(lin_tab):
                    linha = ""
                    # Ano
                    linha += str(int(dsv.tabela[i, 0])).rjust(4) + "  "
                    # Usina
                    linha += str(int(dsv.tabela[i, 1])).rjust(3) + " "
                    # Desvios de cada mês
                    for j in range(n_meses):
                        v = dsv.tabela[i, j + 2]
                        linha += "{:4.2f}".format(v).rjust(6) + " "
                    # Flag de usar desvio
                    linha += str(int(dsv.tabela[i, -1])).rjust(4)
                    arq.write(linha + "\n")

            # Escreve cabeçalhos
            titulos = ("ANO  USIN    JAN    FEV    MAR    ABR    MAI    JUN"
                       + "    JUL    AGO    SET    OUT    NOV    DEZ" + "\n")
            cab = ("XXXX  XXX XXXX.X XXXX.X XXXX.X XXXX.X XXXX.X XXXX.X "
                   + "XXXX.X XXXX.X XXXX.X XXXX.X XXXX.X XXXX.X" + "\n")
            arq.write(titulos)
            arq.write(cab)
            escreve_desvios(dsvagua)
            # Escreve a linha de terminação
            arq.write("9999\n")



# Imports do próprio módulo
from inewave._utils.leitura import Leitura
from inewave.config import MAX_ANOS_ESTUDO, NUM_PATAMARES, MESES
from .modelos.patamar import Patamar
# Imports de módulos externos
import os
import numpy as np  # type: ignore
from traceback import print_exc
from typing import IO, List, Tuple


class LeituraPatamar(Leitura):
    """
    Realiza a leitura do arquivo patamar.dat
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo patamar.dat, construindo
    um objeto `Patamar` cujas informações são as mesmas do patamar.dat.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.

    Uma vez realizada a leitura do arquivo, as informações são guardadas
    internamente no atributo `patamar`.

    **Exemplos**

    >>> diretorio = "~/documentos/.../deck"
    >>> leitor = LeituraPatamar(diretorio)
    >>> leitor.le_arquivo()
    # Ops, esqueci de pegar o objeto
    >>> patamar = leitor.patamar

    """

    str_fim_patamares = "SUBSISTEMA"

    def __init__(self,
                 diretorio: str) -> None:
        super().__init__()
        self.diretorio = diretorio
        # Patamar default, depois é substituído
        self.patamar = Patamar(0, [], np.array([]))

    def le_arquivo(self) -> Patamar:
        """
        Realiza a leitura do arquivo `patamar.dat`.
        """
        try:
            caminho = os.path.join(self.diretorio, "patamar.dat")
            with open(caminho, "r") as arq:
                # Pula inicialmente duas linhas
                # de número de patamares
                self._le_linha_com_backup(arq)
                self._le_linha_com_backup(arq)
                # A terceira linha tem o número de patamares
                linha = self._le_linha_com_backup(arq)
                num_pat = int(linha[1:3])
                if num_pat != NUM_PATAMARES:
                    raise Exception("Número de patamares incompatível")
                # Em seguida, pula três linhas e inicia a leitura
                # da tabela de patamares
                self._le_linha_com_backup(arq)
                self._le_linha_com_backup(arq)
                self._le_linha_com_backup(arq)
                # Lê a tabela de valores
                anos, tabela = self._le_patamares(arq)
                self.patamar = Patamar(NUM_PATAMARES, anos, tabela)
                return self.patamar
        except Exception:
            print_exc()
            return Patamar(0, [], np.array([]))

    def _le_patamares(self, arq: IO) -> Tuple[List[int], np.ndarray]:
        """
        Faz a leitura da tabela de patamares de carga.
        """
        anos: List[int] = []
        patamares = np.zeros((NUM_PATAMARES * MAX_ANOS_ESTUDO,
                              len(MESES)))
        n = 0
        while True:
            # Confere se a tabela já acabou
            linha = self._le_linha_com_backup(arq)
            self._configura_backup()
            if LeituraPatamar.str_fim_patamares in linha:
                # Retorna os anos de estudo e o recorte da tabela
                # com os valores usados
                return anos, patamares[:n*NUM_PATAMARES, :]
            # Senão, lê mais um ano
            a, p = self._le_patamares_ano(arq)
            anos.append(a)
            li = n * NUM_PATAMARES
            lf = li + NUM_PATAMARES
            patamares[li:lf, :] = p
            n += 1

    def _le_patamares_ano(self, arq: IO) -> Tuple[int, np.ndarray]:
        """
        Lê os patamares de um ano na tabela de patamares.
        """
        patamares_ano = np.zeros((NUM_PATAMARES, len(MESES)))
        cols = 6
        ano = 0
        for p in range(NUM_PATAMARES):
            linha = self._le_linha_com_backup(arq)
            if p == 0:
                # Se é o primeiro patamar, extrai o ano
                ano = int(linha[0:4])
            col_i = 6
            for m in range(len(MESES)):
                col_f = col_i + cols
                patamares_ano[p, m] = float(linha[col_i:col_f])
                col_i = col_f + 2
        return ano, patamares_ano

    def _fim_arquivo(self, linha: str) -> bool:
        return False



from typing import Dict
import numpy as np  # type: ignore


class RegistroNwlistcf:
    """
    Armazena as informações de um registro da FCF do NEWAVE.

    ** Parâmetros **

    - ireg: `int`
    - rhs:  `float`
    - tabela: `np.ndarray`

    """

    __slots__ = ["ireg",
                 "rhs",
                 "tabela"]

    def __init__(self,
                 ireg: int,
                 rhs: float,
                 tabela: np.ndarray):
        self.ireg = ireg
        self.rhs = rhs
        self.tabela = tabela

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre RegistroNwlistcf avalia todos os
        valores.
        """
        if not isinstance(o, RegistroNwlistcf):
            return False
        reg: RegistroNwlistcf = o
        eq_ireg = self.ireg == reg.ireg
        eq_rhs = self.rhs == reg.rhs
        eq_tab = np.array_equal(self.tabela, reg.tabela)

        return all([eq_ireg,
                    eq_rhs,
                    eq_tab])


class Nwlistcf:
    """
    Armazena os dados dos cortes construídos pelo NEWAVE existentes
    no arquivo `nwlistcf.rel` do NWLISTCF.

    Esta classe armazena os cortes da FCF de cada uma das variáveis,
    para cada registro e REE dentro do registro.

    Cada registro possui um modelo próprio, armazenando os coeficientes
    dos hiperplanos em uma array específica.

    **Parâmetros**

    - registros: `Dict[int, Dict[int, RegistroNwlistcf]]`

    """
    def __init__(self,
                 registros: Dict[int, Dict[int, RegistroNwlistcf]]):
        self.registros = registros

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre Nwlistcf avalia todos os registros.
        """
        if not isinstance(o, Nwlistcf):
            return False
        nw: Nwlistcf = o
        if self.registros.keys() != nw.registros.keys():
            return False
        for (p1, p2) in zip(self.registros.values(),
                            nw.registros.values()):
            if p1.keys() != p2.keys():
                return False
            for (r1, r2) in zip(p1.values(),
                                p2.values()):
                if r1 != r2:
                    return False
        return True



from inewave.nwlistcf.estados import LeituraEstados

leitor = LeituraEstados("./tests/_arquivos")
leitor.le_arquivo()


def test_leitura():
    assert len(leitor.estados.registros.keys()) > 0


def test_eq_estados():
    leitor2 = LeituraEstados("./tests/_arquivos")
    leitor2.le_arquivo()
    assert leitor2.estados == leitor.estados


def test_neq_estados():
    leitor2 = LeituraEstados("./tests/_arquivos")
    leitor2.le_arquivo()
    leitor2.estados.registros[5] = {}
    assert leitor2.estados != leitor.estados



from inewave.config import REES
from typing import Dict, List
import numpy as np  # type: ignore


class PARp:
    """
    Armazena os dados de saída do NEWAVE referentes aos modelos e às
    séries sintéticas de energia geradas pelo PAR(p) e PAR(p)-A.


    Esta classe lida com informações de saída do NEWAVE e
    cujas saídas devem ser compatíveis com as observadas através
    do NWLISTOP.

    **Parâmetros**

    - ordens_orig: `Dict[int, np.ndarray]`
    - ordens_finais: `Dict[int, np.ndarray]`
    - coeficientes: `Dict[int, np.ndarray]`
    - series_energia: `Dict[int, np.ndarray]`
    - correl_energia: `Dict[int, np.ndarray]`
    - series_medias: `Dict[int, np.ndarray]`
    - correl_media: `Dict[int, np.ndarray]`

    """
    def __init__(self,
                 ordens_orig: Dict[int, np.ndarray],
                 ordens_finais: Dict[int, np.ndarray],
                 coeficientes: Dict[int, np.ndarray],
                 series_energia: Dict[int, np.ndarray],
                 correl_energia: Dict[int, np.ndarray],
                 series_medias: Dict[int, np.ndarray],
                 correl_media: Dict[int, np.ndarray],
                 correl_e_anual: Dict[int, np.ndarray],
                 correl_e_mensal: Dict[int, np.ndarray]):
        self.ordens_orig = ordens_orig
        self.ordens_finais = ordens_finais
        self.coeficientes = coeficientes
        self.series_energia = series_energia
        self.correl_energia = correl_energia
        self.series_medias = series_medias
        self.correl_media = correl_media
        self.correl_e_anual = correl_e_anual
        self.correl_e_mensal = correl_e_mensal

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre PARp avalia todos os campos.
        """
        if not isinstance(o, PARp):
            return False
        parp: PARp = o
        eq_ordens_o = all([np.array_equal(o1, o2)
                           for (o1, o2) in zip(self.ordens_orig.values(),
                                               parp.ordens_orig.values())])
        eq_ordens_f = all([np.array_equal(o1, o2)
                           for (o1, o2) in zip(self.ordens_finais.values(),
                                               parp.ordens_finais.values())])
        eq_coefs = all([np.array_equal(c1, c2)
                        for (c1, c2) in zip(self.coeficientes.values(),
                                            parp.coeficientes.values())])
        eq_series_e = all([np.array_equal(s1, s2)
                           for (s1, s2) in zip(self.series_energia.values(),
                                               parp.series_energia.values())])
        eq_correl_e = all([np.array_equal(s1, s2)
                           for (s1, s2) in zip(self.correl_energia.values(),
                                               parp.correl_energia.values())])
        eq_medias_e = all([np.array_equal(s1, s2)
                           for (s1, s2) in zip(self.series_medias.values(),
                                               parp.series_medias.values())])
        eq_correl_med = all([np.array_equal(s1, s2)
                             for (s1, s2) in zip(self.correl_media.values(),
                                                 parp.correl_media.values())])
        eq_correl_a = all([np.array_equal(s1, s2)
                           for (s1, s2) in zip(self.correl_e_anual.values(),
                                               parp.correl_e_anual.values())])
        eq_correl_m = all([np.array_equal(s1, s2)
                           for (s1, s2) in zip(self.correl_e_mensal.values(),
                                               parp.correl_e_mensal.values())])
        return all([eq_ordens_o,
                    eq_ordens_f,
                    eq_coefs,
                    eq_series_e,
                    eq_correl_e,
                    eq_medias_e,
                    eq_correl_med,
                    eq_correl_a,
                    eq_correl_m])

    def series_energia_ree(self,
                           ree: int) -> Dict[int, np.ndarray]:
        """
        A tabela de séries de energia para todas as configurações
        de uma determinada REE, no mesmo formato do arquivo `parp.dat`,
        excluindo a coluna dos anos.

        **Parâmetros**

        - ree: `int`

        **Retorna**

        `Dict[int, np.ndarray]`

        **Sobre**

        O acesso é feito com [configuracao] e retorna um np.ndarray.
        """
        dict_series: Dict[int, np.ndarray] = {}
        n_configs = self.series_energia[ree].shape[2]
        for c in range(n_configs):
            dict_series[c + 1] = self.series_energia[ree][:, :, c][:, 1:]

        return dict_series

    def series_medias_ree(self,
                          ree: int) -> Dict[int, np.ndarray]:
        """
        A tabela de séries das médias anuais de energia para todos os anos
        de uma determinada REE, no mesmo formato do arquivo `parp.dat`.

        **Parâmetros**

        - ree: `int`

        **Retorna**

        `Dict[int, np.ndarray]`

        **Sobre**

        O acesso é feito com [ano] e retorna um np.ndarray.
        """
        dict_series: Dict[int, np.ndarray] = {}
        n_anos = self.series_medias[ree].shape[2]
        anos = self.anos_estudo
        for c in range(n_anos):
            dict_series[anos[c]] = self.series_medias[ree][:, :, c]

        return dict_series

    def correlograma_energia_ree(self,
                                 ree: int) -> Dict[int,
                                                   np.ndarray]:
        """
        A tabela de autocorrelações parciais da série de energia
        de uma determinada REE, no mesmo formato do arquivo `parp.dat`,
        excluindo a coluna dos meses.

        **Parâmetros**

        - ree: `int`

        **Retorna**

        `Dict[int, np.ndarray]`

        **Sobre**

        O acesso é feito com [mes] e retorna um np.ndarray.
        """
        dict_correl: Dict[int, np.ndarray] = {}
        n_meses = self.correl_energia[ree].shape[0]
        for m in range(n_meses):
            dict_correl[m + 1] = self.correl_energia[ree][m, 1:]

        return dict_correl

    def correlograma_media_ree(self,
                               ree: int) -> Dict[int,
                                                 np.ndarray]:
        """
        A tabela de correlações cruzadas da série de médias anuais de
        energia com as séries de energia de uma determinada REE, no mesmo
        formato do arquivo `parp.dat`, excluindo a coluna dos meses.

        **Parâmetros**

        - ree: `int`

        **Retorna**

        `Dict[int, np.ndarray]`

        **Sobre**

        O acesso é feito com [mes] e retorna um np.ndarray.
        """
        dict_correl: Dict[int, np.ndarray] = {}
        n_meses = self.correl_media[ree].shape[0]
        for m in range(n_meses):
            dict_correl[m + 1] = self.correl_media[ree][m, 1:]

        return dict_correl

    def ordens_originais_ree(self,
                             ree: int) -> Dict[int, np.ndarray]:
        """
        A tabela de ordens originais do modelo PAR ou PAR-A
        de uma determinada REE, no mesmo formato do arquivo `parp.dat`,
        organizada por ano de estudo.

        **Parâmetros**

        - ree: `int`

        **Retorna**

        `Dict[int, np.ndarray]`

        **Sobre**

        O acesso é feito com [ano] e retorna um np.ndarray.
        """
        anos = self.anos_estudo
        ordens = [self.ordens_orig[ree][i, 1:]
                  for i in range(self.ordens_orig[ree].shape[0])]
        return {a: o for a, o in zip(anos, ordens)}

    def ordens_finais_ree(self,
                          ree: int) -> Dict[int, np.ndarray]:
        """
        A tabela de ordens finais do modelo PAR ou PAR-A
        de uma determinada REE, no mesmo formato do arquivo `parp.dat`,
        organizada por ano de estudo.

        **Parâmetros**

        - ree: `int`

        **Retorna**

        `Dict[int, np.ndarray]`

        **Sobre**

        O acesso é feito com [ano] e retorna um np.ndarray.
        """
        anos = self.anos_estudo
        ordens = [self.ordens_finais[ree][i, 1:]
                  for i in range(self.ordens_finais[ree].shape[0])]
        return {a: o for a, o in zip(anos, ordens)}

    def coeficientes_ree(self,
                         ree: int) -> List[np.ndarray]:
        """
        Lista de coeficientes dos modelos PAR ou PAR-A.

        **Parâmetros**

        - ree: `int`

        **Retorna**

        `List[np.ndarray]`

        **Sobre**

        No caso de um modelo PAR-A de ordem p, a lista possui
        p + 1 coeficientes, e a última posição contém o
        coeficiente da componente anual.
        """
        # Extrai os coeficientes regressivos
        todos_coefs = self.coeficientes[ree][:, :, 0]
        coefs: List[List[float]] = []
        for i, c in enumerate(todos_coefs):
            coefs.append([])
            for co in c:
                if co == 0.:
                    break
                coefs[i].append(co)
        # Extrai os coeficientes da componente anual
        coefs_anual = self.coeficientes[ree][:, :, 2]
        for i, c in enumerate(coefs_anual):
            coefs[i].append(c[0])

        return [np.array(c) for c in coefs]

    @property
    def correlacoes_espaciais_anuais(self) -> Dict[int,
                                                   Dict[int,
                                                        Dict[int,
                                                             float]]]:
        """
        Correlações espaciais anuais para cada combinação de
        REEs em cada configuração do sistema, da mesma maneira
        encontrada no arquivo `parp.dat`.

        **Retorna**

        `Dict[int, Dict[int, Dict[int, float]]]`

        **Sobre**

        O acesso é feito com `[configuracao][ree1][ree2]`, onde `ree1`
        e `ree2` são inteiros de 1 ao número de REEs, indexados da mesma
        maneira do NEWAVE (1 = SUDESTE, 2 = SUL...).
        """
        corrs = self.correl_e_anual
        rees = range(1, len(REES) + 1)
        corrs_anuais: Dict[int, Dict[int, Dict[int, float]]] = {}
        for c in corrs.keys():
            if c not in corrs_anuais.keys():
                corrs_anuais[c] = {}
            for ree1 in rees:
                if ree1 not in corrs_anuais[c].keys():
                    corrs_anuais[c][ree1] = {}
                for ree2 in rees:
                    corrs_anuais[c][ree1][ree2] = corrs[c][ree1-1,
                                                           ree2-1]
        return corrs_anuais

    @property
    def correlacoes_espaciais_mensais(self) -> Dict[int,
                                                    Dict[int,
                                                         Dict[int,
                                                              np.ndarray]]]:
        """
        Correlações espaciais mensais para cada combinação de
        REEs em cada configuração do sistema, da mesma maneira
        encontrada no arquivo `parp.dat`.

        **Retorna**

        `Dict[int, Dict[int, Dict[int, np.ndarray]]]`

        **Sobre**

        O acesso é feito com `[configuracao][ree1][ree2]`, onde `ree1`
        e `ree2` são inteiros de 1 ao número de REEs, indexados da mesma
        maneira do NEWAVE (1 = SUDESTE, 2 = SUL...). É retornada a lista
        com as correlações para todos os meses, ordenadas por mês.
        """
        corrs = self.correl_e_mensal
        rees = range(1, len(REES) + 1)
        corrs_mensais: Dict[int, Dict[int, Dict[int, np.ndarray]]] = {}
        for c in corrs.keys():
            if c not in corrs_mensais.keys():
                corrs_mensais[c] = {}
            for ree1 in rees:
                if ree1 not in corrs_mensais[c].keys():
                    corrs_mensais[c][ree1] = {}
                for ree2 in rees:
                    corrs_mensais[c][ree1][ree2] = corrs[c][ree1-1,
                                                            :,
                                                            ree2-1]
        return corrs_mensais

    @property
    def anos_historico(self) -> List[int]:
        """
        A lista de anos do histórico associados às séries de
        energia.

        **Retorna**

        `List[int]`
        """
        a = np.array(self.series_energia[1][:, :, 0][:, 0],
                     dtype=np.int64)
        return list(a)

    @property
    def anos_estudo(self) -> List[int]:
        """
        A lista de anos do estudo associados às tabelas de
        coeficientes e ordens dos modelos.

        **Retorna**

        `List[int]`
        """
        a = np.array(self.ordens_finais[1][:, 0], dtype=np.int64)
        return list(a)



# Rotinas de testes associadas ao arquivo pmo.dat do NEWAVE
from inewave.newave.pmo import LeituraPMO
from inewave.newave.dger import LeituraDGer
from inewave.config import MESES, SUBMERCADOS
import numpy as np  # type: ignore


anos_estudo_teste = [1995, 1996, 1997, 1998, 1999]
leitor = LeituraPMO("tests/_arquivos")
leitor.le_arquivo()


def test_leitura_dados_pmo():
    assert leitor.pmo.ano_pmo == 1995
    assert leitor.pmo.mes_pmo == 5
    assert leitor.pmo.versao_newave == "27.4"


def test_eq_pmo():
    leitor2 = LeituraPMO("tests/_arquivos")
    leitor2.le_arquivo()
    assert leitor.pmo == leitor2.pmo


def test_neq_pmo():
    leitor2 = LeituraPMO("tests/_arquivos")
    leitor2.le_arquivo()
    leitor2.pmo.ano_pmo = 2077
    assert leitor.pmo != leitor2.pmo


def test_anos_estudo():
    assert anos_estudo_teste == leitor.pmo.risco_ens.anos_estudo


def test_leitura_configs():
    configs = leitor.pmo.configuracoes_expansao.configs_por_ano
    assert len(configs.keys()) == 10
    for ano, confs in configs.items():
        assert ano > 0
        assert len(confs) == len(MESES)


def test_leitura_retas_perdas():
    perdas = leitor.pmo.retas_perdas_engolimento
    assert perdas.funcao_perdas(1, -1) < 0


def test_leitura_convergencia():
    convergencia = leitor.pmo.convergencia
    tempos = convergencia.tempos_execucao
    zinfs = convergencia.zinf
    zsups = convergencia.zsup
    assert len(tempos) == 45
    assert len(zinfs.keys())
    assert len(zsups.keys())


def test_leitura_ens():
    risco_ens = leitor.pmo.risco_ens
    por_subs_e_ano = risco_ens.ens_por_subsistema_e_ano
    ano = risco_ens.anos_estudo[-1]
    assert list(por_subs_e_ano.keys()) == SUBMERCADOS
    assert por_subs_e_ano["SUDESTE"][ano] == 0.4
    assert por_subs_e_ano["SUL"][ano] == 0.3


def test_leitura_risco():
    risco_ens = leitor.pmo.risco_ens
    por_subs_e_ano = risco_ens.riscos_por_subsistema_e_ano
    ano = risco_ens.anos_estudo[-1]
    assert list(por_subs_e_ano.keys()) == SUBMERCADOS
    assert por_subs_e_ano["SUDESTE"][ano] == 0.05
    assert por_subs_e_ano["SUL"][ano] == 0.05
    assert por_subs_e_ano["NORTE"][ano] == 0.05


def test_leitura_tabelas_custos():
    custo_series = leitor.pmo.custo_series_simuladas
    valor_esp = leitor.pmo.valor_esperado_periodo
    custo_ref = leitor.pmo.custo_referenciado
    assert np.all(custo_series.custos == valor_esp.custos)
    assert np.all(custo_series.custos == custo_ref.custos)
    assert np.all(custo_ref.custos == valor_esp.custos)


def test_eco_dger_pmo():
    leitor2 = LeituraDGer("tests/_arquivos")
    leitor2.le_arquivo()
    leitor.pmo.dados_gerais.eq_eco_saida(leitor2.dger)



# Rotinas de testes associadas ao arquivo dsvagua.dat do NEWAVE
from inewave.newave.dsvagua import LeituraDSVAgua, EscritaDSVAgua


leitor = LeituraDSVAgua("tests/_arquivos")
leitor.le_arquivo()


def test_leitura():
    assert leitor.dsvagua.tabela.shape[0] > 0


def test_escrita_e_leitura():
    escritor = EscritaDSVAgua("tests/_saidas")
    escritor.escreve_arquivo(leitor.dsvagua)
    leitor2 = LeituraDSVAgua("tests/_saidas")
    leitor2.le_arquivo()
    assert leitor.dsvagua == leitor2.dsvagua


def test_eq_dsvagua():
    leitor2 = LeituraDSVAgua("tests/_arquivos")
    leitor2.le_arquivo()
    assert leitor2.dsvagua == leitor.dsvagua


def test_neq_dsvagua():
    leitor2 = LeituraDSVAgua("tests/_arquivos")
    leitor2.le_arquivo()
    leitor2.dsvagua.tabela[0, 0] = 1e-6
    assert leitor2.dsvagua != leitor.dsvagua
    assert leitor2.dsvagua is not None



# Imports do próprio módulo
from inewave._utils.escrita import Escrita
from inewave._utils.leitura import Leitura
from .modelos.confhd import UHEConfhd, Confhd
# Imports de módulos externos
import os
from traceback import print_exc
from typing import IO, Dict


class LeituraConfhd(Leitura):
    """
    Realiza a leitura do arquivo `confhd.dat`
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo `confhd.dat`, construindo
    um objeto `Confhd` cujas informações são as mesmas do `confhd.dat`.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.

    Uma vez realizada a leitura do arquivo, as informações são guardadas
    internamente no atributo `confhd`.

    **Exemplos**

    >>> diretorio = "~/documentos/.../deck"
    >>> leitor = LeituraConfhd(diretorio)
    >>> leitor.le_arquivo()
    # Ops, esqueci de pegar o objeto
    >>> confhd = leitor.confhd

    """

    def __init__(self,
                 diretorio: str) -> None:
        super().__init__()
        self.diretorio = diretorio
        # Confhd default, depois é substituído
        self.confhd = Confhd({})

    def le_arquivo(self) -> Confhd:
        """
        Faz a leitura do arquivo `confhd.dat`.
        """
        try:
            caminho = os.path.join(self.diretorio, "confhd.dat")
            with open(caminho, "r") as arq:
                self.confhd = self._le_confhd(arq)
                return self.confhd
        except Exception:
            print_exc()
            return self.confhd

    def _le_confhd(self, arq: IO) -> Confhd:
        """
        Faz a leitura do arquivo confhd.dat.
        """
        linha = ""
        # Pula as duas primeiras linhas, com cabeçalhos
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        usinas: Dict[int, UHEConfhd] = {}
        while True:
            # Verifica se o arquivo acabou
            linha = self._le_linha_com_backup(arq)
            if self._fim_arquivo(linha):
                self.confhd = Confhd(usinas)
                break
            # Senão, lê mais uma usina
            u = self._le_dados_uhe(linha)
            usinas[u.numero] = u

        return self.confhd

    def _le_dados_uhe(self, linha: str) -> UHEConfhd:
        """
        Lê a linha com dados do PMO em questão e retorna dados
        sobre o mês e ano de estudo e a versão do NEWAVE usada.
        """
        numero = int(linha[1:5])
        nome = linha[6:18]
        posto = int(linha[19:23])
        jusante = int(linha[25:29])
        ree = int(linha[30:34])
        vol_ini = float(linha[35:41])
        existente = True if linha[42:46].strip() == "EX" else False
        modif = bool(int(linha[49:53]))
        inicio_hist = int(linha[58:62])
        fim_hist = int(linha[67:71])
        return UHEConfhd(numero,
                         nome,
                         posto,
                         jusante,
                         ree,
                         vol_ini,
                         existente,
                         modif,
                         inicio_hist,
                         fim_hist)

    def _fim_arquivo(self, linha: str) -> bool:
        return len(linha) < 1


class EscritaConfhd(Escrita):
    """
    Realiza a escrita do arquivo confhd.dat
    em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para escrever os campos
    de um arquivo confhd.dat, a partir de um objeto `Confhd`.

    Este objeto existe para retirar do modelo de dados a complexidade
    de armazenar as strings auxiliares do arquivo, desenhar tabelas, dentre
    outras tarefas associadas à escrita.

    Se o diretório de escrita não existir, ele será criado.

    **Exemplos**

    >>> diretorio = "~/documentos/.../deck"
    >>> # confhd é do tipo Confhd
    >>> escritor = EscritaConfhd(diretorio)
    >>> escritor.escreve_arquivo(confhd)
    """
    def __init__(self, diretorio: str):
        super().__init__()
        self.diretorio = diretorio

    def escreve_arquivo(self, confhd: Confhd):
        """
        Realiza a escrita de um arquivo `confhd.dat`.
        """
        # Confere se o diretório existe. Senão, cria.
        if not os.path.exists(self.diretorio):
            os.makedirs(self.diretorio)
        # Inicia a escrita
        caminho = os.path.join(self.diretorio, "confhd.dat")
        with open(caminho, "w") as arq:

            def escreve_uhe(uhe: UHEConfhd):
                linha = " "
                # Número
                linha += str(uhe.numero).rjust(4) + " "
                # Nome
                linha += uhe.nome.ljust(12) + " "
                # Posto
                linha += str(uhe.posto).rjust(4) + "  "
                # Jusante
                linha += str(uhe.jusante).rjust(4) + " "
                # REE
                linha += str(uhe.ree).rjust(4) + " "
                # Volume inicial
                linha += "{:3.2f} ".format(uhe.volume_inicial).rjust(7)
                # Existente
                linha += "  EX   " if uhe.existente else "  NE   "
                # Modificada
                linha += str(int(uhe.modificada)).rjust(4) + "     "
                # Início do histórico
                linha += str(uhe.inicio_historico).rjust(4) + "     "
                # Fim do histórico
                linha += str(uhe.fim_historico).rjust(4)
                arq.write(linha + "\n")

            # Escreve cabeçalhos
            titulos = (" NUM  NOME         POSTO JUS   REE V.INIC"
                       + " U.EXIS MODIF INIC.HIST FIM HIST" + "\n")
            cabecalhos = (" XXXX XXXXXXXXXXXX XXXX  XXXX XXXX XXX.XX"
                          + " XXXX   XXXX     XXXX     XXXX" + "\n")
            arq.write(titulos)
            arq.write(cabecalhos)
            # Escreve UHEs
            for uhe in confhd.usinas.values():
                escreve_uhe(uhe)



# Imports do próprio módulo
from inewave._utils.leitura import Leitura
from inewave.config import NUM_CENARIOS
from inewave.config import NUM_PATAMARES, MESES, SUBMERCADOS
from .modelos.cmarg00 import Cmarg00
# Imports de módulos externos
import os
import numpy as np  # type: ignore
from traceback import print_exc
from typing import IO, Dict, Tuple


class LeituraCmarg00(Leitura):
    """
    Realiza a leitura dos arquivos cmarg00x.out
    existentes em um diretório de saídas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de arquivos cmarg00x.out, construindo
    objetos `Cmarg00` cujas informações são as mesmas dos arquivos.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.

    Uma vez realizada a leitura do arquivo, as informações são guardadas
    internamente no atributo `cmargs`.

    **Exemplos**

    >>> diretorio = "~/documentos/.../deck"
    >>> leitor = LeituraCmarg00(diretorio)
    >>> leitor.le_arquivos()
    # Ops, esqueci de pegar o objeto
    >>> cmarg_sudeste = leitor.cmargs['SUDESTE']

    """
    str_inicio_cmargs_ano = "     ANO: "
    str_fim_cmargs_ano = "  MAX         "

    def __init__(self,
                 diretorio: str) -> None:
        super().__init__()
        self.diretorio = diretorio
        self.arquivos = self._lista_arquivos_por_chave("cmarg00")
        self.cmargs: Dict[str, Cmarg00] = {}

    def le_arquivos(self) -> Dict[str, Cmarg00]:
        """
        Lê os arquivos cmarg00x.out em um diretório.
        """
        caminhos = [os.path.join(self.diretorio, f)
                    for f in self.arquivos]
        for a, c in zip(self.arquivos, caminhos):
            cmarg = self._le_arquivo(c)
            self.cmargs[cmarg.submercado] = cmarg

        return self.cmargs

    def _le_arquivo(self, caminho: str) -> Cmarg00:
        """
        Lê um arquivo cmarg00x.out.
        """
        try:
            with open(caminho, "r") as arq:
                # A primeira linha contém as informações gerais
                # da execução do NEWAVE associada
                linha = self._le_linha_com_backup(arq)
                mes, ano, ver = self._infere_informacoes_execucao(linha)
                # Em seguida, infere o submercado e inicia a leitura
                # dos valores
                linha = self._le_linha_com_backup(arq)
                sub = self._infere_submercado(linha)
                # Lê a tabela de valores
                tabelas = self._le_cmarg00(arq)
                cmarg = Cmarg00(mes, ano, ver, sub, tabelas)
                return cmarg
        except Exception:
            print_exc()
            return Cmarg00(0, 0, "", "", {})

    def _infere_informacoes_execucao(self, linha: str) -> Tuple[int,
                                                                int,
                                                                str]:
        """
        """
        # Confere se a linha passada contém as informações.
        # Se não, lança exceção
        if "NW Versao" in linha:
            # Procura o mês de execução
            mes_na_linha = 0
            for i, m in enumerate(MESES):
                if m in linha:
                    mes_na_linha = i
                    break
            mes = MESES[mes_na_linha]
            indice_mes = MESES.index(MESES[mes_na_linha]) + 1
            # Quebra a linha no mês e extrai o ano
            _, linha_sem_mes = linha.split(mes)
            ano = int(linha_sem_mes[3:7])
            # Extrai a versão
            versao = linha.split("NW Versao")[1].strip()
            return indice_mes, ano, versao
        else:
            raise Exception("Linha sem as informações de execução")

    def _infere_submercado(self, linha: str) -> str:
        """
        """
        encontrou = False
        submercado = ""
        for sub in SUBMERCADOS:
            if sub in linha:
                encontrou = True
                submercado = sub
                break
        if encontrou:
            return submercado
        # Se não encontrou o submercado, lança exceção
        raise Exception("Submercado não encontrado")

    def _le_cmarg00(self, arq: IO) -> Dict[int, np.ndarray]:
        """
        Realiza a leitura das tabelas de valores de um arquivo CMarg00.
        """
        iniciou = False
        linha = ""
        tabelas_anos: Dict[int, np.ndarray] = {}
        while True:
            linha = self._le_linha_com_backup(arq)
            # Confere se já leu todos os anos de estudo
            if len(linha) < 1:
                break
            # Procura pelo início da tabela de CMarg do ano
            if not iniciou:
                iniciou = LeituraCmarg00.str_inicio_cmargs_ano in linha
                if iniciou:
                    self._configura_backup()
                continue
            # Extrai o ano da tabela
            ano = int(linha[10:14])
            # Salta uma linha e lê os valores da tabela
            self._le_linha_com_backup(arq)
            tabelas_anos[ano] = self._le_cmarg00_ano(arq)
            # Reseta a flag de início de tabela
            iniciou = False
        return tabelas_anos

    def _le_cmarg00_ano(self, arq: IO) -> np.ndarray:
        """
        Lê os dados da tabela de um ano do arquivo cmarg00x.out.
        """
        n_meses = len(MESES)
        cmargs_ano = np.zeros((NUM_PATAMARES * NUM_CENARIOS, n_meses))
        for c in range(NUM_CENARIOS):
            li = NUM_PATAMARES * c
            lf = li + NUM_PATAMARES
            cmargs_ano[li:lf, :] = self._le_cmarg00_cenario(arq)
        return cmargs_ano

    def _le_cmarg00_cenario(self, arq: IO) -> np.ndarray:
        """
        Lê os dados de um cenário de uma tabela do arquivo cmarg00x.out.
        Retorna um array P x M, onde P é o número de patamares e M é o
        número de meses. O acesso é feito com [p, m].
        """
        n_meses = len(MESES)
        cmargs_cenario = np.zeros((NUM_PATAMARES, n_meses))
        cols = 8
        for p in range(NUM_PATAMARES):
            linha = self._le_linha_com_backup(arq)
            col_i = 15
            for m in range(n_meses):
                col_f = col_i + cols
                cmargs_cenario[p, m] = float(linha[col_i:col_f])
                col_i = col_f + 1
        return cmargs_cenario

    def _fim_arquivo(self, linha: str) -> bool:
        return False



# Imports do próprio módulo
from inewave._utils.escrita import Escrita
from inewave._utils.leitura import Leitura
from .modelos.vazpast import VazPast
from inewave.config import MESES
# Imports de módulos externos
import os
import numpy as np  # type: ignore
from traceback import print_exc
from typing import IO, List


class LeituraVazPast(Leitura):
    """
    Realiza a leitura do arquivo `vazpast.dat`
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo `vazpast.dat`, construindo
    um objeto `VazPast` cujas informações são as mesmas do `vazpast.dat`.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.

    Uma vez realizada a leitura do arquivo, as informações são guardadas
    internamente no atributo `vazpast`.

    **Exemplos**

    >>> diretorio = "~/documentos/.../deck"
    >>> leitor = LeituraVazPast(diretorio)
    >>> leitor.le_arquivo()
    # Ops, esqueci de pegar o objeto
    >>> vazpast = leitor.vazpast

    """

    def __init__(self,
                 diretorio: str) -> None:
        super().__init__()
        self.diretorio = diretorio
        # Confhd default, depois é substituído
        self.vazpast = VazPast(0, 0, [], [], np.array([]))

    def le_arquivo(self) -> VazPast:
        """
        Faz a leitura do arquivo `vazpast.dat`.
        """
        try:
            caminho = os.path.join(self.diretorio, "vazpast.dat")
            with open(caminho, "r") as arq:
                self.confhd = self._le_vazpast(arq)
                return self.vazpast
        except Exception:
            print_exc()
            return self.vazpast

    def _le_vazpast(self, arq: IO) -> VazPast:
        """
        Faz a leitura do arquivo vazpast.dat.
        """
        # Pula as duas primeiras linhas, com cabeçalhos
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        n_meses = len(MESES)
        # Extrai o mês e ano do estudo
        linha = self._le_linha_com_backup(arq)
        str_mes_plan, str_ano_plan = linha.split("ANOPLAN")
        mes_plan = int(str_mes_plan[9:].strip())
        ano_plan = int(str_ano_plan[9:].strip())
        # Lê as informações da tabela
        postos: List[int] = []
        nomes: List[str] = []
        tabela = np.zeros((5000, n_meses))
        i = 0
        while True:
            # Verifica se o arquivo acabou
            linha = self._le_linha_com_backup(arq)
            if self._fim_arquivo(linha):
                self.vazpast = VazPast(mes_plan,
                                       ano_plan,
                                       postos,
                                       nomes,
                                       tabela[:i, :])
                break
            # Senão, lê mais uma linha
            # Posto
            postos.append(int(linha[:5]))
            # Nome
            nomes.append(linha[6:18].strip())
            # Vazoes passadas de cada mês
            ci = 19
            nc = 9
            for m in range(n_meses):
                cf = ci + nc
                valor = float(linha[ci:cf])
                tabela[i, m] = valor
                ci = cf + 1
            i += 1

        return self.vazpast

    def _fim_arquivo(self, linha: str) -> bool:
        return len(linha) < 1


class EscritaVazPast(Escrita):
    """
    Realiza a escrita do arquivo vazpast.dat
    em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para escrever os campos
    de um arquivo vazpast.dat, a partir de um objeto `VazPast`.

    Este objeto existe para retirar do modelo de dados a complexidade
    de armazenar as strings auxiliares do arquivo, desenhar tabelas, dentre
    outras tarefas associadas à escrita.

    Se o diretório de escrita não existir, ele será criado.

    **Exemplos**

    >>> diretorio = "~/documentos/.../deck"
    >>> # vaz é do tipo VazPast
    >>> escritor = EscritaVazPast(diretorio)
    >>> escritor.escreve_arquivo(vaz)
    """
    def __init__(self, diretorio: str):
        super().__init__()
        self.diretorio = diretorio

    def escreve_arquivo(self, vazpast: VazPast):
        """
        Realiza a escrita de um arquivo `vazpast.dat`.
        """
        # Confere se o diretório existe. Senão, cria.
        if not os.path.exists(self.diretorio):
            os.makedirs(self.diretorio)
        # Inicia a escrita
        caminho = os.path.join(self.diretorio, "vazpast.dat")
        n_meses = len(MESES)
        with open(caminho, "w") as arq:

            def escreve_desvios(vaz: VazPast):
                lin_tab = vaz.tabela.shape[0]
                for i in range(lin_tab):
                    linha = " "
                    # Posto
                    linha += str(vaz.postos[i]).rjust(4) + " "
                    # Nome
                    linha += str(vaz.nomes[i]).ljust(12)
                    # Vazoes de cada mês
                    for j in range(n_meses):
                        v = vaz.tabela[i, j]
                        linha += " " + "{:6.2f}".format(v).rjust(9)
                    arq.write(linha + "\n")

            # Escreve cabeçalhos
            dummy = ("********* DUMMY" + "\n")
            arq.write(dummy)
            arq.write(dummy)
            ano_mes = " "
            ano_mes += f"MESPLAN={str(vazpast.mes_planejamento).rjust(13)}"
            ano_mes += f"  ANOPLAN={str(vazpast.ano_planejamento).rjust(13)}"
            arq.write(ano_mes + "\n")
            escreve_desvios(vazpast)



from inewave.nwlistcf.nwlistcf import LeituraNwlistcf

leitor = LeituraNwlistcf("./tests/_arquivos")
leitor.le_arquivo()


def test_leitura():
    assert len(leitor.nwlistcf.registros.keys()) > 0


def test_eq_nwlistcf():
    leitor2 = LeituraNwlistcf("./tests/_arquivos")
    leitor2.le_arquivo()
    assert leitor2.nwlistcf == leitor.nwlistcf


def test_neq_nwlistcf():
    leitor2 = LeituraNwlistcf("./tests/_arquivos")
    leitor2.le_arquivo()
    leitor2.nwlistcf.registros[5] = {}
    assert leitor2.nwlistcf != leitor.nwlistcf
